"""
This module defines the Assembler class, which is used to convert a Qsearch-style circuit into other formats, such as Qiskit or Qasm.

The DictionaryAssembler subclass is provided as the default implementation of an Assembler.  Use it as-is or as an example for writing your own Assembler.
Some constants are also defined as DictionaryAssembler instances preloaded with the most common assembly dictionaries.

Attributes:
    ASSEMBLER_QISKIT : Outputs Python code that generates a Qiskit circuit object.
    ASSEMBLER_OPENQASM : Outputs generic Openqasm code.  This may not be compatible with IBM Qiskit.
    ASSEMBLER_IBMOPENQASM : Outputs Openqasm code with the IBM imports and gate names.  This flavor of Openqasm is compatible with IBM Qiskit.
    ASSEMBLER_QUTRIT : Outputs pseudocode for circuits built with single-qutrit gates and CNOTs.
"""

from .options import Options

class Assembler():
    """This class is used to translate Qsearch-style circuits to other formats."""

    def __init__(self, options=Options()):
        self.options = options

    def assemble(self, resultdict, options=None):
        """
        The assemble function is used to convert the circuit described in resultdict.  See DictionaryAssembler for an example implementation.

        Args:
            resultdict : The dictionary representing the desired circuit.  It is expected to contain the entries "stucture" and "parameters".  It may contain other entries.

        Returns:
            str: A string representing the converted circuit code.
        """
        raise NotImplementedError()

class DictionaryAssembler(Assembler):
    """This subclass of Assembler uses a dictionary that specifies mappings from gate names to output code, as well as an output code initial line.

        Options:
        assemblydict (required) : A dictionary that specifies mappings from gate names to output code.
    """

    def assemble(self, resultdict, options=None):
        options = self.options.updated(options)
        options.make_required("assemblydict")

        circuit, v = resultdict["structure"], resultdict["parameters"]
        il = flatten_intermediate(circuit.assemble(v))
        assemblydict = options.assemblydict

        out = assemblydict["initial"].format(circuit.qudits)

        for segment in il:
            if not segment[0] == "gate":
                continue
            try:
                parsed = assemblydict[segment[1]].format(*segment[2], *segment[3])
            except:
                raise KeyError("Unable to compile gate {} to the language defined by{}".format(segment, assemblydict))
            out += parsed

        return out


def flatten_intermediate(intermediate):
    """This is a helper function for working with the intermediate tuple language that is output by the assemble method of QuantumStep objects."""
    flattened = []
    for step in intermediate:
        if step[0] == "gate":
            flattened.append(step)
        elif step[0] == "block":
            flattened += flatten_intermediate(step[1])
        else:
            raise KeyError("Found unexpected tuple type {} in assembly tuple {}".format(step[0], step))
    return flattened





# These are example dictionaries for use with DictionaryAssembler.  The can be modified or used as examples for writing your own such dictionaries.

assemblydict_qiskit = {
        # A dictionary that allows DictionaryAssembler to export Python code to generate a Qiskit circuit
        "initial" : "num_qubits = {}\nqr = QuantumRegister(num_qubits)\ncr = ClassicalRegister(num_qubits)\nqc = QuantumCircuit(qr,cr)\n",
        "X" : "qc.rx({}, {})\n",
        "Y" : "qc.ry({}, {})\n",
        "Z" : "qc.rz({}, {})\n",
        "U3" : "qc.u({}, {}, {}, {})\n",
        "CNOT" : "qc.cx({}, {})\n",
}

# Currently ibm's quantum experience cannot handle normal openqasm, even though supposedly u3 and cx below should evaluate to U3 and CX above.
# This pure openqasm does not seem to be supported by IBM.  I have double checked that the syntax is theoretically correct.  The IBM Quantum Experience circuit builder will only recognize the ibm versions.
assemblydict_openqasm = {
        # A dictionary that allows DictionaryAssembler to export code to generate an Openqasm script.
        "initial" : 'OPENQASM 2.0;\nqreg q[{}];\n',
        "X" : "U({}, -pi/2, pi/2) q[{}];\n",
        "Y" : "U({}, 0, 0) q[{}];\n",
        "Z" : "U(0, 0, {}) q[{}];\n",
        "U3" : "U({}, {}, {}) q[{}];\n",
        "CNOT" : "CX q[{}], q[{}];\n"
}

# currently ibm's quantum experience cannot handle normal openqasm, even though supposedly u3 and cx below should evaluate to U3 and CX above.
assemblydict_ibmopenqasm = {
        # A dictionary that allows DictionaryAssembler to export code to generate an Openqasm script that is compatible with IBM Qiskit.
        "initial" : 'OPENQASM 2.0;\ninclude \"qelib1.inc\";\n\nqreg q[{}];\n',
        "X" : "rx({}) q[{}];\n",
        "Y" : "ry({}) q[{}];\n",
        "Z" : "rz({}) q[{}];\n",
        "U3" : "u3({}, {}, {}) q[{}];\n",
        "CNOT" : "cx q[{}], q[{}];\n"
}

assemblydict_qutrit = {
        # A dictionary that allows DictionaryAssembler to export a human-readable pseudocode representation of a circuit using single qutrit gates and CNOTs.
        "initial" : "",
        "QUTRIT" : "qutrit({}, {}, {}, {}, {}, {}, {}, {}) index: {}\n",
        "CNOT" : "cnot control: {} target: {}\n"
}


# Some constants for the most common use cases.  These are instances of Assembler subclasses, and can be passed unmodified to Project or used on a Qsearch-style circuit.
ASSEMBLER_QISKIT = DictionaryAssembler(Options(assemblydict=assemblydict_qiskit))
ASSEMBLER_OPENQASM = DictionaryAssembler(Options(assemblydict=assemblydict_openqasm))
ASSEMBLER_IBMOPENQASM = DictionaryAssembler(Options(assemblydict=assemblydict_ibmopenqasm))
ASSEMBLER_QUTRIT = DictionaryAssembler(Options(assemblydict=assemblydict_qutrit))



"""
This module defines solvers that use multiple starting points in order to have a higher chance at finding the global minimum.
"""
from . import utils, logging
from .solvers import Solver, default_solver
import numpy as np
import scipy as sp
import scipy.optimize
from qsrs import native_from_object
import time
from math import pi, gamma, sqrt

import multiprocessing as mp
import sys
from .persistent_aposmm import initialize_APOSMM, decide_where_to_start_localopt, update_history_dist, add_to_local_H

def distance_for_x(x, options, circuit):
    """Calculate the distance between circuit and the target for input x based on the distance metric"""
    if options.inner_solver.distance_metric == "Frobenius":
        return options.error_func(options.target, circuit.matrix(x))
    elif options.inner_solver.distance_metric == "Residuals":
        return np.sum(options.error_residuals(options.target, circuit.matrix(x), np.eye(options.target.shape[0]))**2)

def optimize_worker(circuit, options, q, x0):
    """Worker function used to run the inner solver in parallel"""
    _, xopt = options.inner_solver.solve_for_unitary(circuit, options, x0)
    q.put((distance_for_x(xopt, options, circuit), xopt))

class MultiStart_Solver(Solver):
    """A higher accuracy solver based on APOSMM https://www.mcs.anl.gov/~jlarson/APOSMM/

    MultiStart_Solver generally gets better results than other optimizers due to the advanced algorithm
    to start multiple local optimizers ("inner solvers") and find the global optimum more often.
    """

    def __init__(self, num_threads):
        """Create a MultiStart_Solver instance. Pass num_threads to set how many threads to use in parallel optimization runs"""
        self.num_threads = num_threads
        self.ctx = mp.get_context('fork') if sys.platform != 'win32' else mp.get_context()

    def solve_for_unitary(self, circuit, options, x0=None):
        """Optimize the given circuit based on the provided options with initial point x0 (optional).
      
        Args:
            circuit: A qsearch.gates.Gate describing the circuit to optimize
            options: This uses the following options:
                    - inner_solver : which optimizer to use for local optimization runs
                    - target : the target unitary of synthesis
                    - logger : A qsearch.logging.Logger that will be used for logging the synthesis process.
                    - error_func : The function that the Solver will attempt to minimize.
                    - error_residuals : A function that returns an array of real-valued residuals to be used by a least-squares-based Solver.

            x0: the starting point for the optimzier
        """
        if 'inner_solver' not in options:
            options.inner_solver = default_solver(options)
        U = options.target
        logger = options.logger if "logger" in options else logging.Logger(verbosity=options.verbosity, stdout_enabled=options.stdout_enabled, output_file=options.log_file)

        #np.random.seed(4) # usually we do not want fixed seeds, but it can be useful for some debugging
        n = circuit.num_inputs # the number of parameters to optimize (the length that v should be when passed to one of the lambdas created above)
        initial_sample_size = 100  # How many points do you want to sample before deciding where to start runs.
        num_localopt_runs = self.num_threads  # How many localopt runs to start?

        specs = {'lb': np.zeros(n),
                 'ub': np.ones(n),
                 'standalone': True,
                 'initial_sample_size':initial_sample_size}

        _, _, rk_const, ld, mu, nu, _, H = initialize_APOSMM([],specs,None)

        initial_sample = np.random.uniform(0, 1, (initial_sample_size, n))

        add_to_local_H(H, initial_sample, specs, on_cube=True)

        for i, x in enumerate(initial_sample):
            H['f'][i] = distance_for_x(2*np.pi*x, options, circuit)

        H[['returned']] = True

        update_history_dist(H, n)
        starting_inds = decide_where_to_start_localopt(H, n, initial_sample_size, rk_const, ld, mu, nu)

        starting_points = H['x'][starting_inds[:num_localopt_runs]]

        start = time.time()
        q = self.ctx.Queue()
        processes = []
        rets = []
        for x0 in starting_points:
            p = self.ctx.Process(target=optimize_worker, args=(circuit, options, q, 2*np.pi*x0))
            processes.append(p)
            p.start()
        for p in processes:
            ret = q.get() # will block
            rets.append(ret)
        for p in processes:
            p.join()
        end = time.time()

        best_found = np.argmin([r[0] for r in rets])
        best_val = rets[best_found][0]

        xopt = rets[best_found][1]

        return (circuit.matrix(xopt), xopt)

class NaiveMultiStart_Solver(Solver):
    """A naive but effective multi-start solver which tries to cover as much of the optimization space at once"""
    def __init__(self, num_threads):
        """Create a NaiveMultiStart_Solver instance. Pass num_threads to set how many threads to use in parallel optimization runs"""
        self.threads = num_threads if num_threads else 1
        self.ctx = mp.get_context('fork') if sys.platform != 'win32' else mp.get_context()

    def solve_for_unitary(self, circuit, options, x0=None):
        if 'inner_solver' not in options:
            options.inner_solver = default_solver(options)
        U = options.target
        logger = options.logger if "logger" in options else logging.Logger(verbosity=options.verbosity, stdout_enabled=options.stdout_enabled, output_file=options.log_file)
        n = circuit.num_inputs
        initial_samples = [np.random.uniform((i - 1)/self.threads, i/self.threads, (circuit.num_inputs,)) for i in range(1, self.threads+1)]
        q = self.ctx.Queue()
        processes = []
        rets = []
        for x0 in initial_samples:
            p = self.ctx.Process(target=optimize_worker, args=(circuit, options, q, x0))
            processes.append(p)
            p.start()
        for p in processes:
            ret = q.get() # will block
            rets.append(ret)
        for p in processes:
            p.join()

        best_found = np.argmin([r[0] for r in rets])
        best_val = rets[best_found][0]

        xopt = rets[best_found][1]

        return (circuit.matrix(xopt), xopt)



"""
Defines Solver, a class used to wrap various numerical optimizers for finding parameters such that an ansatz circuit is a solution to a target unitary.
"""
import sys

import numpy as np
import scipy as sp
import scipy.optimize

from . import utils
from .gatesets import *
from .logging import Logger
try:
    from qsrs import LeastSquares_Jac_SolverNative, BFGS_Jac_SolverNative, native_from_object, matrix_residuals, matrix_residuals_jac
except ImportError:
    LeastSquares_Jac_SolverNative = BFGS_Jac_SolverNative = native_from_object = matrix_residuals= matrix_residuals_jac = None

def default_solver(options, x0=None):
    """Runs a complex list of tests to determine the best Solver for a specific situation."""
    options = options.copy()
    # re-route the default behavior for error_func and error_jac because the default functions for those parameters often rely on the return valye from default_solver
    options.set_defaults(logger=Logger(), U=np.array([]), error_func=None, error_jac=None, target=None)
    options.remove_smart_defaults("error_func", "error_jac")

    # Choosse the best default solver for the given gateset
    ls_failed = False

    # check if Rust works on the layers
    gateset = options.gateset
    qudits = 0 if options.target is None else int(np.log(options.target.shape[0]) // np.log(gateset.d))

    rs_failed = True
    if native_from_object is not None:
        layers = [(gateset.initial_layer(qudits), 0)] + gateset.search_layers(qudits)
        for layer in layers:
            try:
                native_from_object(layer[0])
            except ValueError:
                break
        else:
            rs_failed = False

    # Check to see if the gateset and error func are explicitly supported by LeastSquares
    error_func = options.error_func
    error_jac = options.error_jac
    logger = options.logger

    if type(gateset).__module__ != QubitCNOTLinear.__module__:
        ls_failed = True
    elif error_func is not None and (error_func.__module__ != utils.matrix_distance_squared.__module__ or (error_func.__name__ != utils.matrix_distance_squared.__name__ and error_func.__name__ != utils.matrix_residuals.__name__)):
        ls_failed = True

    if not ls_failed:
        # since all provided gatesets support jacobians, this is the only check we need
        if rs_failed or options.error_residuals not in (utils.matrix_residuals, matrix_residuals) or options.error_residuals_jac not in (utils.matrix_residuals_jac, matrix_residuals_jac):
            logger.logprint("Smart default chose LeastSquares_Jac_Solver", verbosity=3)
            return LeastSquares_Jac_Solver()
        else:
            logger.logprint("Smart default chose LeastSquares_Jac_SolverNative", verbosity=3)
            return LeastSquares_Jac_SolverNative()

    if qudits < 1:
        logger.logprint("Smart default fell back to COBYLA_Solver.  Pass a different Solver to SearchCompiler for better results.", verbosity=1)
        return COBYLA_Solver() # handling this case for manually created SearchCompiler instances.  Better support for manual usage is unlikely to be implemented because Projects are generally recommended.

    # least squares won't work, so check for jacobian and rust success
    jac_failed = False
    for layer, _ in layers:
        try:
            layer.mat_jac(np.random.rand(layer.num_inputs))
        except:
            jac_failed = True

    if error_jac is None and error_func not in [None, utils.matrix_distance_squared, utils.matrix_residuals]:
        jac_failed = True

    if jac_failed:
        logger.logprint("Smart default chose COBYLA_Solver", verbosity=2)
        return COBYLA_Solver()
    else:
        logger.logprint("Smart default chose BFGS_Jac_Solver", verbosity=2)
        return BFGS_Jac_Solver()
    # the default will have been chosen from LeastSquares, BFGS, or COBYLA

class Solver():
    """This class is used to wrap numerical optimizers for circuit solving."""
    def solve_for_unitary(self, circuit, options, x0=None):
        """Finds the best parameters that minimize error_func or error_residuals between the unitary from the circuit and options.target."""
        raise NotImplementedError

    def __eq__(self, other):
        if self is other:
            return True
        if self.__module__ == Solver.__module__:
            if type(self) == type(other):
                return True
        return False

    @property
    def distance_metric(self):
        return "Frobenius"

class CMA_Solver(Solver):
    """Uses cmaes gradient-free optimization from the cma package."""
    def solve_for_unitary(self, circuit, options, x0=None):
        try:
            import cma
        except ImportError:
            print("ERROR: Could not find cma, try running pip install quantum_synthesis[cma]", file=sys.stderr)
            sys.exit(1)
        eval_func = lambda v: options.error_func(options.target, circuit.matrix(v))
        initial_guess = 'np.random.rand({})*2*np.pi'.format(circuit.num_inputs) if x0 is None else x0
        xopt, _ = cma.fmin2(eval_func, initial_guess, 0.25, {'verb_disp':0, 'verb_log':0, 'bounds' : [0,2*np.pi]}, restarts=2)
        return (circuit.matrix(xopt), xopt)

class COBYLA_Solver(Solver):
    """Uses cobyla gradient-free optimization from scipy."""
    def solve_for_unitary(self, circuit, options, x0=None):
        eval_func = lambda v: options.error_func(options.target, circuit.matrix(v))
        initial_guess = np.array(np.random.rand(circuit.num_inputs))*2*np.pi if x0 is None else x0
        x = sp.optimize.fmin_cobyla(eval_func, initial_guess, cons=[lambda x: np.all(np.less_equal(x,2*np.pi))], rhobeg=0.5, rhoend=1e-12, maxfun=1000*circuit.num_inputs)
        return (circuit.matrix(x), x)

class DIY_Solver(Solver):
    """An easier way to wrap a numerical optimizer than writing your own Solver class."""
    def __init__(self, f):
        """Uses the function f that takes in eval_func and initial_guess and returns the parameters that minimizes eval_func."""
        self.f = f

    def solve_for_unitary(self, circuit, options, x0=None):
        eval_func = lambda v: options.error_func(options.target, circuit.matrix(v))
        initial_guess = np.array(np.random.rand(circuit.num_inputs))*2*np.pi if x0 is None else x0
        x = f(eval_func, initial_guess)

class NM_Solver(Solver):
    """A solver based on the Nelder-Mead gradient free optimizer from scipy."""
    def solve_for_unitary(self, circuit, options, x0=None):
        eval_func = lambda v: options.error_func(options.target, circuit.matrix(v))
        result = sp.optimize.minimize(eval_func, np.random.rand(circuit.num_inputs)*2*np.pi if x0 is None else x0, method='Nelder-Mead', options={"ftol":1e-14})
        xopt = result.x
        return (circuit.matrix(xopt), xopt)

class CMA_Jac_Solver(Solver):
    """A solver based on the cmaes optimizer from the cma package, but using gradients."""
    def solve_for_unitary(self, circuit, options, x0=None):
        try:
            import cma
        except ImportError:
            print("ERROR: Could not find cma, try running pip install quantum_synthesis[cma]", file=sys.stderr)
            sys.exit(1)
        eval_func = lambda v: options.error_func(options.target, circuit.matrix(v))
        jac_func  = lambda v: options.error_jac(options.target, circuit.mat_jac(v))
        initial_guess = 'np.random.rand({})'.format(circuit.num_inputs)*2*np.pi if x0 is None else x0
        xopt, es = cma.fmin2(eval_func, initial_guess, 0.25, {'verb_disp':0, 'verb_log':0, 'bounds' : [0,2*np.pi]}, restarts=2, gradf=jac_func)
        if circuit.num_inputs > 18:
            raise Warning("Finished with {} evaluations".format(es.result[3]))
        return (circuit.matrix(xopt), xopt)

class BFGS_Jac_Solver(Solver):
    """A solver based on the BFGS implementation in scipy.  It requires gradients."""
    def solve_for_unitary(self, circuit, options, x0=None):
        def eval_func(v):
            M, jacs = circuit.mat_jac(v)
            return options.error_jac(options.target, M, jacs)
        result = sp.optimize.minimize(eval_func, np.random.rand(circuit.num_inputs)*2*np.pi if x0 is None else x0, method='BFGS', jac=True)
        xopt = result.x
        return (circuit.matrix(xopt), xopt)

class LeastSquares_Jac_Solver(Solver):
    """Uses the Leavenberg-Marquardt least-squares optimizer in scipy."""
    def solve_for_unitary(self, circuit, options, x0=None):
        # This solver is usually faster than BFGS, but has some caveats
        # 1. This solver relies on matrix residuals, and therefore ignores the specified error_func, making it currently not suitable for alternative synthesis goals like stateprep
        # 2. This solver (currently) does not correct for an overall phase, and so may not be able to find a solution for some gates with some gatesets.  It has been tested and works fine with QubitCNOTLinear, so any single-qubit and CNOT-based gateset is likely to work fine.
        I = np.eye(options.target.shape[0])
        eval_func = lambda v: options.error_residuals(options.target, circuit.matrix(v), I)
        jac_func = lambda v: options.error_residuals_jac(options.target, *circuit.mat_jac(v))
        if options.max_quality_optimization:
            result = sp.optimize.least_squares(eval_func, np.random.rand(circuit.num_inputs)*2*np.pi if x0 is None else x0, jac_func, method="lm", ftol=5e-16, xtol=5e-16, gtol=1e-15)
        else:
            result = sp.optimize.least_squares(eval_func, np.random.rand(circuit.num_inputs)*2*np.pi if x0 is None else x0, jac_func, method="lm")
        xopt = result.x
        return (circuit.matrix(xopt), xopt)

    @property
    def distance_metric(self):
        return "Residuals"




"""
This module provides defaults for Options objects.  This includes definitions of smart_default functions, and dictionaries to be used with set_defaults and set_smart_defaults.

Three default dictionaries are provided.

Attributes:
    standard_defaults : A dictionary containing defaults for standard gate synthesis.
    standard_smart_defaults : A dictionary containing smart_defaults functions for standard gate synthesis.
    stateprep_defaults : A dictionary containing defaults for stateprep synthesis.
"""

from . import utils, gatesets, solvers, backends, parallelizers, heuristics, logging, checkpoints, assemblers
from functools import partial



def default_eval_func(options):
    if options.error_func == utils.matrix_residuals:
        return utils.matrix_distance_squared
    else:
        return options.error_func

def default_heuristic(options):
    if options.search_type == "astar":
        return heuristics.astar
    if options.search_type == "djikstra":
        return heuristics.djikstra
    elif options.search_type == "greedy":
        return heuristics.greedy
    raise KeyError("Unknown search_type {}, and no alternative heuristic provided.".format(options.search_type))

def default_error_jac(options):
    if options.error_func == utils.matrix_distance_squared:
        return utils.matrix_distance_squared_jac
    else:
        return None

def default_error_residuals_jac(options):
    if options.error_residuals == utils.matrix_residuals:
        return utils.matrix_residuals_jac
    else:
        return None

def default_logger(options):
    return logging.Logger(verbosity=options.verbosity, stdout_enabled=options.stdout_enabled, output_file=options.log_file)

def default_checkpoint(options):
    return checkpoints.FileCheckpoint(options=options)

def identity(U):
    return U

standard_defaults = {
        "threshold":1e-10,
        "gateset":gatesets.Default(),
        "beams":-1,
        "delta": 0,
        "weight_limit":None,
        "search_type":"astar",
        "statefile":None,
        "error_func":utils.matrix_distance_squared,
        "error_residuals":utils.matrix_residuals,
        "backend":backends.SmartDefaultBackend(),
        "parallelizer":parallelizers.MultiprocessingParallelizer,
        "log_file":None,
        "max_quality_optimization" : False,
        "assembler" : assemblers.ASSEMBLER_QISKIT,
        "write_location" : None,
        "unitary_preprocessor": utils.nearest_unitary,
        "timeout" : float('inf'),
        }
standard_smart_defaults = {
        "eval_func":default_eval_func,
        "error_jac":default_error_jac,
        "error_residuals_jac":default_error_residuals_jac,
        "solver":solvers.default_solver,
        "heuristic":default_heuristic,
        "logger" :default_logger,
        "checkpoint":default_checkpoint,
        }

stateprep_defaults = {
        "error_residuals" : partial(utils.matrix_residuals_slice, (0, slice(None))),
        "error_residuals_jac" : partial(utils.matrix_residuals_slice_jac, (0, slice(None))),
        "eval_func" : partial(utils.eval_func_from_residuals, partial(utils.matrix_residuals_slice, (0, slice(None)))),
        "unitary_preprocessor": identity
        }




"""
A class for holding and managing options that are passed to various other classes in the Qsearch suite.

Options objects work like dictionaries, but if an Options object is queried for an item and it does not have it, it first checks its defaults and smart_defaults properties for the item before throwing an error.  This allows the setting of default values which are easily overridden by user-provided values.  The smart_defaults dictionary contains functions that return an object, allowing for default behavior that is dependent on other settings within the Options object.

Options objects are also designed to be easily combinable through functions such as update and updated.

Options objects are used ubiquitously throughout Qsearch
"""

import pickle
import os

_options_actual_parameters = ["defaults", "smart_defaults", "required", "cache", "load_error"]

class Options():
    """This class manages options that are passed between various Qsearch objects."""
    def __init__(self, defaults={}, smart_defaults={}, **xtraargs):
        self.defaults = dict()
        self.smart_defaults = dict()
        self.required = set()
        self.cache = dict()
        self.load_error = None
        self.update(**xtraargs)
        self.set_defaults(**defaults)
        self.set_smart_defaults(**smart_defaults)
        


    def filtered(self, *names):
        """Returns an Options object with only parameters in the specified list names."""
        new_dict = {name:self.__dict__[name] for name in names if name in self.__dict__ and not name in _options_actual_parameters}
        new_defaults = {name:self.defaults[name] for name in names if name in self.defaults}
        new_smart_defaults = {name:self.smart_defaults[name] for name in names if name in self.smart_defaults}
        
        newOptions = Options(**new_defaults)
        newOptions.__dict__.update(new_dict)
        newOptions.smart_defaults = new_smart_defaults
        return newOptions

    def __getitem__(self, name):
        return getattr(self, name)

    def __delitem__(self, name):
        if name in _options_actual_parameters:
            raise AttributeError("This is one of the essential parameters and cannot be removed")
        del self.__dict__[name]

    def __getattr__(self, name):
        if name in _options_actual_parameters:
            raise AttributeError("This options class is corrupt as it is missing its critical members")
        elif name in self.required:
            raise AttributeError("Missing required parameter {}".format(name))
        elif name in self.cache:
            return self.cache[name]
        elif name in self.smart_defaults:
            # this behavior of removing the smart default function and putting it back is intended
            # to generate an error rather than recurse infinitely when there are interdependent smart default functions
            smartfunc = self.smart_defaults[name]
            self.required.add(name)
            retval = smartfunc(self)
            self.smart_defaults[name] = smartfunc
            self.cache[name] = retval
            self.required.remove(name)
            return retval
        elif name in self.defaults:
            return self.defaults[name]
        raise AttributeError("Could not find option {}".format(name))

    def __setattr__(self, name, value):
        if name not in _options_actual_parameters:
            self.cache = dict()
        super().__setattr__(name, value)

    def __contains__(self, name):
        if name in self.__dict__:
            return True
        elif name in self.defaults:
            return True
        elif name in self.smart_defaults:
            return True
        else:
            return False

    def empty_copy(self):
        """Create an Options object with the same defaults but without any specific values."""
        newOptions = Options(**self.defaults)
        newOptions.smart_defaults = self.smart_defaults
        newOptions.required = self.required.copy()
        return newOptions

    def copy(self):
        """Create a full copy of an Options object."""
        newOptions = Options(**self.defaults)
        newOptions.smart_defaults.update(self.smart_defaults)
        newOptions._update_dict(self.__dict__)
        newOptions.required = self.required.copy()
        newOptions.cache = self.cache.copy()
        return newOptions

    def __copy__(self):
        newOptions = Options(**self.defaults)
        newOptions.smart_defaults.update(self.smart_defaults)
        newOptions._update_dict(self.__dict__)
        newOptions.required = self.required.copy()
        newOptions.cache = self.cache.copy()
        return newOptions

    def updated(self, other=None, **xtraargs):
        """Return a new Options object that is a copy of this object, updated with the contents of other and xtraargs."""
        newOptions = self.copy()
        if other is not None:
            newOptions._update_dict(other.__dict__)
            newOptions.defaults.update(other.defaults)
            newOptions.smart_defaults.update(other.smart_defaults)
            newOptions.required = self.required.union(other.required)
        newOptions._update_dict(xtraargs)
        return newOptions

    def update(self, other=None, **xtraargs):
        """Mutate the current Options object with the contents of other and xtraargs."""
        if other is not None:
            self._update_dict(other.__dict__)
            self.defaults.update(other.defaults)
            self.smart_defaults.update(other.smart_defaults)
            self.required.update(other.required)
        self._update_dict(xtraargs)
        self.cache = dict()


    def _update_dict(self, otherdict):
        for name in otherdict:
            if name in _options_actual_parameters:
                continue
            self.__dict__[name] = otherdict[name]

    def set_defaults(self, **args):
        """
        Set default values for this Options object.
        
        If an Options object is queried for a value, and it does not contain it, it will check its defaults list before throwing an error.
        """
        self.defaults.update(args)
        for name in args:
            if name in self.smart_defaults:
                del self.smart_defaults[name]
        self.cache = dict()

    # if a smart_default function is specified, it will be called instead of falling back to normal defaults
    # smart default functions should take an Options class and return a value
    def set_smart_defaults(self, **args):
        """
        Set smart_defaults values for this Options object.

        If an Options object is queried for a value, and it does not contain it, it will check its smart_defaults list before throwing an error.  If it does find a function in smart_defaults, it calls that function, passing itself as the argument, and returns the return value of that function, caching it for next time.
        """
        self.smart_defaults.update(args)
        for name in args:
            if name in self.defaults:
                del self.defaults[name]
        self.cache = dict()

    def make_required(self, *names):
        """Marking names as required will cause the Options object to throw an error if it does not contain it, even if it has defaults or smart_defaults defined."""
        self.required.update(names)
        self.cache = dict()

    def remove_defaults(self, *names):
        """Removes the defaults for the specified names."""
        for name in names:
            if name in self.defaults:
                del self.defaults[name]
        self.cache = dict()

    def remove_smart_defaults(self, *names):
        """Removes the smart_defaults for the specified names."""
        for name in names:
            if name in self.smart_defaults:
                del self.smart_defaults[name]
        self.cache = dict()

    def generate_cache(self):
        """Caches valuesa for all functions in smart_defaults."""
        for key in self.smart_defaults:
            getattr(self, key) 
            # this may look weird, but calling getattr on these keys will populate
            # the cache with any keys that have not been already cached.

    def save(self, filepath=None):
        """Saves the Options object to a file, or to a returned tuple."""
        main_dict = dict()
        for name in self.__dict__:
            if not name in _options_actual_parameters:
                main_dict[name] = pickle.dumps(self.__dict__[name])

        defaults_dict = dict()
        for name in self.defaults:
            if not name in _options_actual_parameters:
                defaults_dict[name] = pickle.dumps(self.defaults[name])

        smart_defaults_dict = dict()
        for name in self.smart_defaults:
            if not name in _options_actual_parameters:
                smart_defaults_dict[name] = pickle.dumps(self.smart_defaults[name])

        if filepath is None:
            return (main_dict, defaults_dict, smart_defaults_dict)
        pickle.dump((main_dict, defaults_dict, smart_defaults_dict), filepath)


    def load(self, filepath_or_tuple, strict=False):
        """
        Loads the Options object from a file or tuple.
        
        If strict is left as False, the Options object will attempt to gracefully handle errors when loading its contents, relying on its ability to fallback to defaults or smart_defaults if those are able to load successfully.

        If strict is set to True, the Options object will throw an error upon any error while loading.
        """
        try:
            if type(filepath_or_tuple) is tuple:
                main_dict, defaults_dict, smart_defaults_dict = filepath_or_tuple
            else:
                main_dict, defaults_dict, smart_defaults_dict = pickle.load(filepath)
        except Exception as e:
            if strict:
                raise
            else:
                self.load_error = e
            return

        for name in main_dict:
            try:
                self.__dict__[name] = pickle.loads(main_dict[name])
            except Exception as e:
                if strict:
                    raise
                else:
                    self.load_error = e

        for name in defaults_dict:
            try:
                self.defaults[name] = pickle.loads(defaults_dict[name])
            except Exception as e:
                if strict:
                    raise
                else:
                    self.load_error = e

        for name in smart_defaults_dict:
            try:
                self.smart_defaults[name] = pickle.loads(smart_defaults_dict[name])
            except Exception as e:
                if strict:
                    raise
                else:
                    self.load_error = e

    def __getstate__(self):
        return self.save()

    def __setstate__(self, state):
        self.defaults = dict()
        self.smart_defaults = dict()
        self.required = set()
        self.cache = dict()
        self.load_error = None

        try:
            self.load(state, strict=True)
        except Exception as e:
            self.load(state, strict=False)
            self.load_error = e




"""
This module defines PostProcessor, a class used to modify circuits after they have been synthesized.

Several implementations are provided.

Attributes:
    BasicSingleQubitReduction_PostProcessor : Attempts to remove single-qubit gates without sacrificing the quality of the solution in terms of eval_func
    ParameterTuning_PostProcessor : Attempts to reduce eval_func simply by re-running the solver with stronger parameters.
    LEAPReoptimizing_PostProcessor : Reduces the length of circuits produced using LEAP by re-running segments of the circuit.
"""
from . import options as opt
from functools import partial
from timeit import default_timer as timer
import heapq
from scipy.stats import linregress
import numpy as np

from .gates import *

from . import solvers as scsolver
from .options import Options
from .defaults import standard_defaults as defaults, standard_smart_defaults as smart_defaults
from . import parallelizers, backends
from . import utils, heuristics, gates, logging, gatesets
from .compiler import Compiler, SearchCompiler
from .checkpoints import ChildCheckpoint

class PostProcessor():
    """This class is used to modify circuits that have already been synthesized."""
    def __init__(self, options = opt.Options()):
        self.options=options

    def post_process_circuit(self, result, options=None):
        """
        Processes the circuit dictionary and returns a new one.

        Args:
            result : A dictionary containing a synthesized circuit.  Expect it to contain "structure" and "parameters", but it may contain more, depending on what previous PostProcessors were run and on the compiler.

        Returns:
            dict : A dictionary containing any updates that should be made to the circuit dictionary, such as new values for "structure" or "parameters" or arbitrary other data.    
        """
        return result


class BasicSingleQubitReduction_PostProcessor(PostProcessor):
    """Attempts to reduce the number of single-qubit gates in a circuit by sequentially removing a gate, attempting to use a Solver on it, and keeping that gate removed if successful."""
    def post_process_circuit(self, result, options=None):
        circuit = result["structure"]
        finalx = result["parameters"]
        options = self.options.updated(options)
        if "unitary_preprocessor" in options:
            target = options.unitary_preprocessor(options.target)
        else:
            target = options.target
        single_qubit_names = ["U3Gate()", "ZXZXZGate()", "XZXZGate()"]
        identitystr = "IdentityGate()"

        circstr = repr(circuit)
        initial_count = sum([circstr.count(sqn) for sqn in single_qubit_names])
        options.logger.logprint("Initial count: {}".format(initial_count), verbosity=2)
        finalcirc = circuit
        for gate in single_qubit_names:
            components = circstr.split(gate)
            while len(components) > 1:
                newstr = components[0] + identitystr + "".join([component + gate for component in components[1:-1]]) + components[-1]
                newcirc = eval(newstr)
                mat, xopt = options.solver.solve_for_unitary(newcirc, options)
                if options.eval_func(target, mat) < options.threshold:
                    components = [components[0] + identitystr + components[1]] + components[2:]
                    finalx = xopt
                    finalcirc = newcirc
                else:
                    components = [components[0] + gate + components[1]] + components[2:]

            circstr = components[0]
        finalstr = repr(finalcirc)
        final_count = sum([finalstr.count(sqn) for sqn in single_qubit_names])
        options.logger.logprint("Final count: {}".format(final_count), verbosity=2)
        options.logger.logprint("Post-processing removed {}, or {}% of the single qubit gates".format(initial_count-final_count, 100*(initial_count-final_count)/initial_count))
        return {"structure":finalcirc, "parameters":finalx}

class ParameterTuning_PostProcessor(PostProcessor):
    """Attempts to reduce the eval_func value of the circuit simply by tuning the parameters better using stronger Solver parameters."""
    def post_process_circuit(self, result, options=None):
        circuit = result["structure"]
        initialx = result["parameters"]
        options = self.options.updated(options)
        options.max_quality_optimization = True
        if "unitary_preprocessor" in options:
            target = options.unitary_preprocessor(options.target)
        else:
            target = options.target
        initial_value = options.eval_func(target, circuit.matrix(initialx))
        options.logger.logprint("Initial Distance: {}".format(initial_value))

        U, x = options.solver.solve_for_unitary(circuit, options)

        final_value = options.eval_func(target, U)
        if np.abs(final_value) < np.abs(initial_value):
            options.logger.logprint("Improved Distance: {}".format(final_value))
            return {"parameters":x}
        else:
            options.logger.logprint("Rejected Distance: {}".format(final_value))
            return {}

class LEAPReoptimizing_PostProcessor(Compiler, PostProcessor):
    """A PostProcessor that re-optimizes LeapCompiler-compiled circuits via search.

    This PostProcessor puts "holes" in the circuit where LEAP fixed prefixes and runs
    qsearch on those holes to reduce the total number of gates.
    """
    def __init__(self, options=Options()):
        self.options = options.copy()
        self.options.set_defaults(verbosity=1, logfile=None, stdout_enabled=True, **defaults)
        self.options.set_smart_defaults(**smart_defaults)

    def post_process_circuit(self, result, options=None):
        """Re-optimize a LEAP circuit. Pass "depth" to indicate the size to re-synthesize.
        It is recommended to call like:
        `project.post_process(post_processing.LEAPReoptimizing_PostProcessor(), solver=multistart_solvers.MultiStart_Solver(8), parallelizer=parallelizers.ProcessPoolParallelizer, depth=7)`
        """
        if len(result['structure']._subgates) <= options.weight_limit if 'weight_limit' in options and options.weight_limit else options.reoptimize_size:
            return result
        best_pair = (result['structure'], result['parameters'])
        opts = options.updated(best_pair=best_pair, cut_depths=result['cut_depths'])
        return self.compile(opts)


    def compile(self, options=Options()):
        """Backwards compatible interface since this is technically a Compiler.

        You should use LEAPReoptimizing_PostProcessor.post_process_circuit with the Project post_processing API.
        """
        options = self.options.updated(options)
        options.make_required("target")

        if "unitary_preprocessor" in options:
            U = options.unitary_preprocessor(options.target)
        depth = options.weight_limit if 'weight_limit' in options else options.reoptimize_size
        child_checkpoint = ChildCheckpoint(Options(parent=options.checkpoint))

        logger = options.logger if "logger" in options else logging.Logger(verbosity=options.verbosity, stdout_enabled=options.stdout_enabled, output_file=options.log_file)

        overall_startime = timer() # note, because all of this setup gets included in the total time, stopping and restarting the project may lead to time durations that are not representative of the runtime under normal conditions
        qudits = int(np.round(np.log(np.shape(U)[0])/np.log(options.gateset.d)))

        parallel = options.parallelizer(options)
        recovered_outer = child_checkpoint.recover_parent()
        if recovered_outer is None:
            overall_best_pair = options.best_pair
            start_depth = len(overall_best_pair[0]._subgates) - 1
            if 'cut_depths' in options:
                # these are the "ideal" starting points, but we may need to modify them as we optimize
                midpoints = [1] + [pt + int((pt - prev)/2) for pt, prev in zip(options.cut_depths[1:],options.cut_depths)]
                logger.logprint(f'Midpoints initialized as {midpoints}', verbosity=2)
            start_point = 1
            overall_best_value = options.eval_func(U, overall_best_pair[0].matrix(overall_best_pair[1]))
        else:
            overall_best_pair, start_depth, midpoints, start_point, overall_best_value = recovered_outer
        while True:
            if 'timeout' in options and timer() - overall_startime > options.timeout:
                break
            best_circuit = overall_best_pair[0]
            best_circuit_depth = len(best_circuit._subgates) - 1
            if 'cut_depths' in options:
                insertion_points = midpoints
            else:
                insertion_points = range(start_point,best_circuit_depth)
            for point in insertion_points:
                if 'timeout' in options and timer() - overall_startime > options.timeout:
                    break
                startime = timer() # note, because all of this setup gets included in the total time, stopping and restarting the project may lead to time durations that are not representative of the runtime under normal conditions
                window_size = depth or options.reoptimize_size
                root = ProductGate(*best_circuit._subgates[:point], *best_circuit._subgates[point + window_size:])
                h = options.heuristic
                qudits = int(np.round(np.log(np.shape(U)[0])/np.log(options.gateset.d)))

                if options.gateset.d**qudits != np.shape(U)[0]:
                    raise ValueError("The target matrix of size {} is not compatible with qudits of size {}.".format(np.shape(U)[0], self.options.gateset.d))

                I = gates.IdentityGate(d=options.gateset.d)

                initial_layer = options.initial_layer if 'initial_layer' in options else options.gateset.initial_layer(qudits)
                search_layers = options.gateset.search_layers(qudits)

                if len(search_layers) <= 0:
                    logger.logprint("This gateset has no branching factor so only an initial optimization will be run.")
                    root = initial_layer
                    result = options.solver.solve_for_unitary(options.backend.prepare_circuit(root, options), options)
                    return (root, result[1])

                # TODO move these print statements somewhere else
                # this is good informati
                logger.logprint("There are {} processors available to Pool.".format(options.num_tasks))
                logger.logprint("The branching factor is {}.".format(len(search_layers)))
                beams = int(options.beams)
                if beams < 1 and len(search_layers) > 0:
                    beams = int(options.num_tasks // len(search_layers))
                if beams < 1:
                    beams = 1
                if beams > 1:
                    logger.logprint("The beam factor is {}.".format(beams))

                recovered_state = child_checkpoint.recover()
                queue = []
                best_depth = 0
                best_value = 0
                best_pair  = 0
                tiebreaker = 0
                rectime = 0
                if recovered_state == None:
                    result = options.solver.solve_for_unitary(options.backend.prepare_circuit(root, options), options)
                    best_value = options.eval_func(U, result[0])
                    best_pair = (root, result[1])
                    logger.logprint("New best! {} at depth 0".format(best_value))
                    if depth == 0:
                        return best_pair

                    queue = [(h(*best_pair, 0, options), 0, best_value, -1, result[1], root)]
                    #         heuristic      depth  distance tiebreaker parameters structure
                    #             0            1      2         3         4        5
                    child_checkpoint.save((queue, best_depth, best_value, best_pair, tiebreaker, timer()-startime))
                else:
                    queue, best_depth, best_value, best_pair, tiebreaker, rectime = recovered_state
                    logger.logprint("Recovered state with best result {} at depth {}".format(best_value, best_depth))

                options.generate_cache() # cache the results of smart_default settings, such as the default solver, before entering the main loop where the options will get pickled and the smart_default functions called many times because later caching won't persist cause of pickeling and multiple processes

                while len(queue) > 0:
                    if 'timeout' in options and timer() - overall_startime > options.timeout:
                        break
                    if best_value < options.threshold:
                        queue = []
                        break
                    popped = []
                    for _ in range(0, beams):
                        if len(queue) == 0:
                            break
                        tup = heapq.heappop(queue)
                        popped.append(tup)
                        logger.logprint("Popped a node with score: {} at depth: {}".format((tup[2]), tup[1]), verbosity=2)

                    then = timer()
                    new_steps = [(current_tup[5].inserting(search_layer[0], depth=point), current_tup[1], search_layer[1]) for search_layer in search_layers for current_tup in popped]
                    for step, result, current_depth, weight in parallel.solve_circuits_parallel(new_steps):
                        current_value = options.eval_func(U, result[0])
                        new_depth = current_depth + weight
                        if (current_value < best_value and (best_value >= options.threshold or new_depth <= best_depth)) or (current_value < options.threshold and new_depth < best_depth):
                            best_value = current_value
                            best_pair = (step, result[1])
                            best_depth = new_depth
                            logger.logprint("New best! score: {} at depth: {}".format(best_value, new_depth))
                        if depth is None or new_depth < depth - 1:
                            heapq.heappush(queue, (h(step, result[1], new_depth, options), new_depth, current_value, tiebreaker, result[1], step))
                            tiebreaker+=1
                    logger.logprint("Layer completed after {} seconds".format(timer() - then), verbosity=2)
                    if (options.weight_limit is not None and best_depth >= options.weight_limit - 1) or ('reoptimize_size' in options and best_depth >= options.reoptimize_size - 1):
                        break
                    child_checkpoint.save((queue, best_depth, best_value, best_pair, tiebreaker, rectime+(timer()-startime)))


                logger.logprint("Finished compilation at depth {} with score {} after {} seconds.".format(best_depth, best_value, rectime+(timer()-startime)))
                new_circuit_depth = len(best_pair[0]._subgates) - 1
                if best_value < options.threshold and new_circuit_depth < best_circuit_depth:
                    logger.logprint(f"With starting point {point} re-optimized from depth {best_circuit_depth} to depth {new_circuit_depth}")
                    overall_best_pair = best_pair
                    overall_best_value = best_value
                    # select the points which are greater than the search window and adjust by new reoptimization
                    logger.logprint(f'old midpoints: {midpoints}')
                    midpoints = [i - (best_circuit_depth - new_circuit_depth) for i in midpoints if (i - (point + window_size)) > 0]
                    logger.logprint(f'new midpoints: {midpoints}')
                    child_checkpoint.save(None)
                    child_checkpoint.save_parent((overall_best_pair, start_depth, midpoints, start_point, overall_best_value))
                    break # break out so we can re-run optimization on the better circuit
                else:
                    logger.logprint(f"With starting point {point} no improvement was made to depth", verbosity=2)
                    logger.logprint(f'old midpoints: {midpoints}')
                    midpoints = [i for i in midpoints if (i - (point + window_size)) > 0]
                    logger.logprint(f'new midpoints: {midpoints}')
                    start_point = point
                    child_checkpoint.save(None)
                    child_checkpoint.save_parent((overall_best_pair, start_depth, midpoints, start_point, overall_best_value))
                    continue
            if new_circuit_depth >= best_circuit_depth:
                break
        parallel.done()
        logger.logprint("Finished all compilations at depth {} with score {} after {} seconds.".format(best_circuit_depth, overall_best_value, rectime+(timer()-overall_startime)))
        return {'structure': overall_best_pair[0], 'parameters': overall_best_pair[1]}



"""
This module provides LeapCompiler, which is a more scalable variant of SearchCompiler, at the expense of producing somewhat longer circuits.  LeapReoptimizing_PostProcessor can be used to reduce circuit length back to levels that SearchCompiler might generate.
"""
from functools import partial
from timeit import default_timer as timer
import heapq
from scipy.stats import linregress
import numpy as np

from . import solvers as scsolver
from .options import Options
from .defaults import standard_defaults, standard_smart_defaults
from . import parallelizers, backends
from . import utils, heuristics, gates, logging, gatesets
from .compiler import Compiler, SearchCompiler
from .checkpoints import ChildCheckpoint


def cut_end(circ, depth):
    if isinstance(circ._substeps[0], gates.ProductGate):
        return cut_end(circ._substeps[0], depth)
    return gates.ProductGate(*circ._substeps[:-depth])


class LeapCompiler(Compiler):
    """LeapCompiler is a more scalable search based circuit compiler

    LeapCompiler uses fixed structure prefixes to greatly reduce the search space
    and speed up synthesis at the cost of optimiality. Thus it is recommended to use in conjunction
    with reoptimizing_compiler.LeapReoptimizing_PostProcessor() to obtain the best results.

    Options:
        target (required) : The unitary matrix to be synthesized, in the form of a numpy ndarray with dtype="complex128".
        gateset : The Gateset used for synthesis.
        weight_limit : A limit on the maximum weight for circuits to be expanded for further searching.  See gatesets.py for more information.  The default is None for unlimited.
        heuristic : A heuristic used to order the search tree.  See heuristics.py for more information.
        solver : A Solver used for optimizing the parameters in parameterized circuits generated by the search tree.
        parallelizer : A Parallelizer used for solving multiple parameterized circuits in parallel.
        beams : The number of nodes to pop from the search tree at a time.  The default value of -1 will create enough branches to maximize utilization of your CPU.
        error_func : The function that the Solver will attempt to minimize.
        eval_func : The function used by the heuristic in order to guide the search tree.  By default this is equal to error_func.
        error_jac : A function that returns a tuple of the value that error_func would generate and the jacobian of error_func
        error_residuals : A function that returns an array of real-valued residuals to be used by a least-squares-based Solver.
        error_residuals_jac : A function that returns the jacobian of error_residuals (note that it does NOT return a tuple of the residuals and the jacobian).
        timeout : An uper limit on the amount of time the compiler will spend trying to synthesize a circuit.  The default is float('inf'), for unlimited.
        checkpoint : The compiler will use this Checkpoint to save intermediate state, and will resume from this Checkpoint if there was an existing state.
        logger : A qsearch.logging.Logger that will be used for logging the synthesis process.
        min_depth : the minimum amount of searching 
    """
    def __init__(self, options=Options()):
        """Run LEAP on the compilation specified in options.
        
        Args:
            options: options for the compilations, see the class level documentation for details.
        """
        self.options = options.copy()
        self.options.set_defaults(verbosity=1, logfile=None, stdout_enabled=True, **standard_defaults)
        self.options.set_smart_defaults(**standard_smart_defaults)

    def compile(self, options=Options()):
        """Run LEAP on the compilation specified in options.
        
        Args:
            options: options for the compilations, see the class level documentation for details.
        """

        options = self.options.updated(options)
        options.make_required("target")

        U = options.unitary_preprocessor(options.target)
        depth = options.weight_limit

        child_checkpoint = ChildCheckpoint(Options(parent=options.checkpoint))

        logger = options.logger if "logger" in options else logging.Logger(verbosity=options.verbosity, stdout_enabled=options.stdout_enabled, output_file=options.log_file)

        starttime = timer() # note, because all of this setup gets included in the total time, stopping and restarting the project may lead to time durations that are not representative of the runtime under normal conditions
        rectime = 0
        qudits = int(np.round(np.log(np.shape(U)[0])/np.log(options.gateset.d)))

        sub_compiler = options.sub_compiler_class if 'sub_compiler_class' in options else SubCompiler
        sc = sub_compiler(options)
        recovered_state = child_checkpoint.recover_parent()
        if recovered_state is None:
            total_depth = 0
            best_value = 1.0
            depths = [total_depth]
            initial_layer = options.gateset.initial_layer(qudits)
        else:
            total_depth = recovered_state[0]
            best_value = recovered_state[1]
            depths = recovered_state[2]
            rectime = recovered_state[3]
            initial_layer = recovered_state[4]
        while True:
            if 'timeout' in options and timer() - starttime > options.timeout:
                break
            opts = options.updated(initial_layer=initial_layer, local_threshold=options.delta * best_value, overall_starttime=starttime, overall_best_value=best_value, checkpoint=child_checkpoint)
            best_pair, best_value, best_depth = sc.compile(opts)
            # clear child checkpoint for next run
            child_checkpoint.delete()
            total_depth += best_depth
            depths.append(total_depth)
            if best_value < options.threshold:
                break
            initial_layer = best_pair[0]
            child_checkpoint.save_parent((total_depth, best_value, depths, timer()-starttime, initial_layer))
        logger.logprint("Finished all sub-compilations at depth {} with score {} after {} seconds.".format(total_depth, best_value, (timer()-starttime)))
        return {'structure': best_pair[0], 'parameters': best_pair[1], 'cut_depths': depths}


class SubCompiler(Compiler):
    """A modified SearchCompiler for the LeapCompiler to use.
    """
    def __init__(self, options=Options()):
        self.options = options.copy()
        self.options.set_defaults(verbosity=1, logfile=None, stdout_enabled=True, **standard_defaults)
        self.options.set_smart_defaults(**standard_smart_defaults)

    def compile(self, options=Options()):
        options = self.options.updated(options)
        options.make_required("target")

        if "unitary_preprocessor" in options:
            U = options.unitary_preprocessor(options.target)
        depth = options.weight_limit
        checkpoint = options.checkpoint

        logger = options.logger if "logger" in options else logging.Logger(verbosity=options.verbosity, stdout_enabled=options.stdout_enabled, output_file=options.log_file)

        starttime = timer() # note, because all of this setup gets included in the total time, stopping and restarting the project may lead to time durations that are not representative of the runtime under normal conditions
        h = options.heuristic
        qudits = int(np.round(np.log(np.shape(U)[0])/np.log(options.gateset.d)))

        if options.gateset.d**qudits != np.shape(U)[0]:
            raise ValueError("The target matrix of size {} is not compatible with qudits of size {}.".format(np.shape(U)[0], self.options.gateset.d))

        I = gates.IdentityGate(d=options.gateset.d)

        initial_layer = options.initial_layer if 'initial_layer' in options else options.gateset.initial_layer(qudits)
        search_layers = options.gateset.search_layers(qudits)

        if len(search_layers) <= 0:
            logger.logprint("This gateset has no branching factor so only an initial optimization will be run.")
            root = initial_layer
            result = options.solver.solve_for_unitary(options.backend.prepare_circuit(root, options), options)
            return (root, result[1])

        parallel = options.parallelizer(options)
        # TODO move these print statements somewhere else
        # this is good informati
        logger.logprint("There are {} processors available to Pool.".format(options.num_tasks))
        logger.logprint("The branching factor is {}.".format(len(search_layers)))
        beams = int(options.beams)
        if beams < 1 and len(search_layers) > 0:
            beams = int(options.num_tasks // len(search_layers))
        if beams < 1:
            beams = 1
        if beams > 1:
            logger.logprint("The beam factor is {}.".format(beams))

        recovered_state = checkpoint.recover()
        queue = []
        best_depth = 0
        best_value = 0
        best_pair  = 0
        tiebreaker = 0
        rectime = 0
        if recovered_state == None:
            if isinstance(initial_layer, gates.ProductGate):
                root = initial_layer
            else:
                root = gates.ProductGate(initial_layer)
            result = options.solver.solve_for_unitary(options.backend.prepare_circuit(root, options), options)
            best_value = options.eval_func(U, result[0])
            best_pair = (root, result[1])
            logger.logprint("New best! {} at depth 0".format(best_value))
            if depth == 0:
                return (best_pair, best_value, 0)

            queue = [(h(*best_pair, 0, options), 0, best_value, -1, result[1], root)]
            #         heuristic      depth  distance tiebreaker parameters structure
            #             0            1      2         3         4        5
            checkpoint.save((queue, best_depth, best_value, best_pair, tiebreaker, timer()-starttime))
        else:
            queue, best_depth, best_value, best_pair, tiebreaker, rectime = recovered_state
            logger.logprint("Recovered state with best result {} at depth {}".format(best_value, best_depth))

        options.generate_cache() # cache the results of smart_default settings, such as the default solver, before entering the main loop where the options will get pickled and the smart_default functions called many times because later caching won't persist cause of pickeling and multiple processes
        previous_bests_depths = []
        previous_bests_values = []
        while len(queue) > 0:
            if 'timeout' in options and timer() - options.overall_starttime > options.timeout:
                break
            if best_value < options.threshold:
                queue = []
                break
            popped = []
            for _ in range(0, beams):
                if len(queue) == 0:
                    break
                tup = heapq.heappop(queue)
                popped.append(tup)
                logger.logprint("Popped a node with score: {} at depth: {}".format((tup[2]), tup[1]), verbosity=2)

            then = timer()
            new_steps = [(current_tup[5].appending(search_layer[0]), current_tup[1], search_layer[1]) for search_layer in search_layers for current_tup in popped]
            for step, result, current_depth, weight in parallel.solve_circuits_parallel(new_steps):
                current_value = options.eval_func(U, result[0])
                new_depth = current_depth + weight
                if (current_value < best_value and (best_value >= options.threshold or new_depth <= best_depth)) or (current_value < options.threshold and new_depth < best_depth):
                    best_value = current_value
                    best_pair = (step, result[1])
                    best_depth = new_depth
                    logger.logprint("New best! score: {} at depth: {}".format(best_value, new_depth))
                    if len(previous_bests_values) > 1:
                        with np.errstate(invalid='ignore', divide='ignore'):
                            slope, intercept, _rval, _pval, _stderr = linregress(previous_bests_depths, previous_bests_values)
                        predicted_best = slope * new_depth + intercept
                        delta = predicted_best - best_value
                        logger.logprint(f"Predicted best value {predicted_best} for new best with delta {delta}", verbosity=2)
                        if not np.isnan(predicted_best) and best_value < options.overall_best_value and delta < 0 and ('min_depth' not in options or new_depth >= options.min_depth):
                            parallel.done()
                            return (best_pair, best_value, best_depth)
                    previous_bests_depths.append(best_depth)
                    previous_bests_values.append(best_value)

                if depth is None or new_depth < depth:
                    heapq.heappush(queue, (h(step, result[1], new_depth, options), new_depth, current_value, tiebreaker, result[1], step))
                    tiebreaker+=1
            logger.logprint("Layer completed after {} seconds".format(timer() - then), verbosity=2)
            checkpoint.save((queue, best_depth, best_value, best_pair, tiebreaker, rectime+(timer()-starttime)))


        logger.logprint("Finished compilation at depth {} with score {} after {} seconds.".format(best_depth, best_value, rectime+(timer()-starttime)))
        parallel.done()
        return (best_pair, best_value, best_depth)



"""
This module contains miscellaneous helper functions and tools.

The functions you may want to be aware of:

Attributes:
    endian_reverse : Reverses the endianness of the specified unitary.  Necessary for working with unitaries from Qiskit.
    matrix_distance_squared : The default error_func.  Returns the Hilbert-Schmidt norm between two matrices.
    matrix_distance_squared_jac : Returns the value that matrix_distance_squared would return, as well as the jacobian.
    matrix_residuals : The default error_residuals.  Returns residuals based on difference between the poduct of the implemented matrix and the hermitian conjugate of the target and the identitiy.
    matrix_residuals_jac : Returns the jacobian of matrix_residuals.  Does not return the value of matrix_residuals as well.
    remap : Remaps a unitary for acting on qudits in a different order.
    upgrade_qudits : Upgrades a unitary from a lower qudit size to a larger qudit size.
"""
import numpy as np
import scipy as sp
import scipy.linalg

try:
    from mpi4py import MPI
except ImportError:
    MPI = None

from . import unitaries

def matrix_product(*LU):
    """Performs matrix multiplication of a list of matrices."""
    result = np.array(np.eye(LU[0].shape[0]), dtype='complex128')
    for U in LU:
        result = np.matmul(U, result, out=result)
    return result

def matrix_kron(*LU):
    """Performs the kronecker product on a list of matrices."""
    result = LU[0]
    for U in LU[1:]:
        result = np.kron(result,U)
    return result

def op_norm(A):
    """An implementation of the l1-l1 operator norm."""
    return max([np.linalg.norm(x,ord=1) for x in A])

def matrix_distance_squared(A,B):
    # this distance function is designed to be phase agnostic
    # optimized implementation
    return 1 - np.abs(np.sum(np.multiply(A,np.conj(B)))) / A.shape[0]
    #original implementation
    #return 1 - np.abs(np.trace(np.dot(A,B.T.conjugate()))) / A.shape[0]

def matrix_distance_squared_jac(U, M, J):
    S = np.sum(np.multiply(U, np.conj(M)))
    dsq = 1 - np.abs(S)/U.shape[0]
    JU = np.array([np.multiply(U,np.conj(K)) for K in J])
    JUS = np.sum(JU, axis=(1,2))
    jacs = -(np.real(S)*np.real(JUS) + np.imag(S)*np.imag(JUS))*U.shape[0] / np.abs(S)
    return (dsq, jacs)

def matrix_residuals(A, B, I):
    M = np.dot(B,np.conj(A.T))
    #M *= np.abs(M[0][0])/M[0][0]
    Re, Im = np.real(M), np.imag(M)
    Re -= I
    Re = np.reshape(Re, (1,-1))
    Im = np.reshape(Im, (1,-1))
    return np.append(Re, Im)

def matrix_residuals_jac(U, M, J):
    Ut = np.conj(U.T)
    #M2 = np.dot(M, Ut)
    #Ut *= np.abs(M2[0][0])/M2[0][0]
    JU = [np.matmul(K, Ut) for K in J]
    JU = np.array([np.append(np.reshape(np.real(K), (1,-1)), np.reshape(np.imag(K), (1,-1))) for K in JU])
    return JU.T

def matrix_residuals_v2(A, B, I):
    # TODO examine how this function behaves compared to the original implementation.  Faster?  Behaves better?  Roughly the same?  Exactly the same?
    M = B - A
    #M *= np.abs(M[0][0])/M[0][0]
    Re, Im = np.real(M), np.imag(M)
    Re = np.reshape(Re, (1,-1))
    Im = np.reshape(Im, (1,-1))
    return np.append(Re, Im)

def matrix_residuals_v2_jac(U, M, J):
    JU = np.array([np.append(np.reshape(np.real(K), (1,-1)), np.reshape(np.imag(K), (1,-1))) for K in J])
    return JU.T

def matrix_residuals_slice(slices, A, B, I):
    M = (B - A)[slices]
    #M *= np.abs(M[0][0])/M[0][0]
    Re, Im = np.real(M), np.imag(M)
    Re = np.reshape(Re, (1,-1))
    Im = np.reshape(Im, (1,-1))
    return np.append(Re, Im)

def matrix_residuals_slice_jac(slices, A, B, J):
    JU = np.array([np.append(np.reshape(np.real(K[slices]), (1,-1)), np.reshape(np.imag(K[slices]), (1,-1))) for K in J])
    return JU.T

def matrix_residuals_blacklist(badrows, badcols, A, B, I):
    M = (B - A)
    M = np.delete(M, badrows, 0)
    M = np.delete(M, badcols, 1)
    #M *= np.abs(M[0][0])/M[0][0]
    Re, Im = np.real(M), np.imag(M)
    Re = np.reshape(Re, (1,-1))
    Im = np.reshape(Im, (1,-1))
    return np.append(Re, Im)

def matrix_residuals_blacklist_jac(slices, A, B, J):
    JU = np.array([np.append(np.reshape(np.real(np.delete(np.delete(K, badrows, 0), badcols, 1)), (1,-1)), np.reshape(np.imag(np.delete(np.delete(K, badrows, 0), badcols, 1)), (1,-1))) for K in J])
    return JU.T

def nearest_unitary(A):
    """
    Calculate the closest unitary to a given matrix.

    Calculate the unitary matrix U that is closest with respect to the
    operator norm distance to the general matrix A.

    D.M.Reich. "Characterisation and Identification of Unitary Dynamics
    Maps in Terms of Their Action on Density Matrices"

    Args:
        A (np.ndarray): The matrix input.

    Returns:
        (np.ndarray): The unitary matrix closest to A.
        Return U as a numpy matrix.

    Thank you to Ed Younis, this is based on code from qfast
    """
    try:
        if len(A.shape) == 2 and A.shape[0] != A.shape[1]:
            raise TypeError("A must be a square matrix.")

        V, __, Wh = sp.linalg.svd(A)
        U = np.array(V.dot(Wh))
        return U
    except Exception:
        return A

def index_test(i, di, df):
    if i < df:
        return False
    elif i % di >= df:
        return True
    else:
        return index_test(i//di)

def downgrade_qudits_residuals(di, df, A, B, I):
    M = (B - A)
    qudits = int(np.log(U.shape[0])/np.log(di))
    M = np.delete(M, [i for i in range(di**qudits) if index_test(i, di, df)], axis=0)
    M = np.delete(M, [i for i in range(di**qudits) if index_test(i, di, df)], axis=1)
    #M *= np.abs(M[0][0])/M[0][0]
    Re, Im = np.real(M), np.imag(M)
    Re = np.reshape(Re, (1,-1))
    Im = np.reshape(Im, (1,-1))
    return np.append(Re, Im)

def downgrade_qudits_residuals_jac(di, df, A, B, J):
    JU = [np.delete(K, [i for i in range(di**qudits) if index_test(i, di, df)], axis=0) for K in J]
    JU = [np.delete(K, [i for i in range(di**qudits) if index_test(i, di, df)], axis=1) for K in JU]
    JU = np.array([np.append(np.reshape(np.real(K), (1,-1)), np.reshape(np.imag(K), (1,-1))) for K in JU])
    return JU.T

def eval_func_from_residuals(f, A, B):
    return np.sum(np.square(f(A,B,I=np.array(np.eye(A.shape[0]), dtype='float64'))))

def generate_stateprep_target_matrix(state):
    # WARNING: the matrix generated by this function (currently) is not unitary, and therefore should only be used with functions like matrix_residuals_slice
    state = np.array(state, dtype='complex128')
    M = np.diag(state)
    M[0] = state
    return M

def matrix_distance(A,B):
    return np.sqrt(np.abs(matrix_distance_squared(A,B)))

def re_rot_z(theta, old_z):
    old_z[0,0] = np.exp(-1j*theta/2)
    old_z[1,1] = np.exp(1j*theta/2)

def re_rot_z_jac(theta, old_z, multiplier=1):
    old_z[0,0] = multiplier*0.5*(-np.sin(theta/2)-1j*np.cos(theta/2))
    old_z[1,1] = multiplier*0.5*(-np.sin(theta/2)+1j*np.cos(theta/2))


# copied from qnl_analysis.SimTools
def q1_unitary(x):
    return matrix_product(rot_z(x[0]), rot_x(np.pi/2), rot_z(np.pi + x[1]), rot_x(np.pi/2), rot_z(x[2] - np.pi))

def qt_arb_rot(Theta_1, Theta_2, Theta_3, Phi_1, Phi_2, Phi_3, Phi_4, Phi_5):
    # this code is now deprecated, because SingleQutritStep now contains a better implementation (more compact and more performant)
    # make sure to move this red comment over to SingleQutritStep before deleting this code for documentation purposes
    """Using the parameterization found in https://journals.aps.org/prd/pdf/10.1103/PhysRevD.38.1994,
        this method constructs an arbitrary single_qutrit unitary operation.

        Arguments:
            qutrit_params: a list of eight parameters, in the following order
                Theta_1, Theta_2, Theta_3, Phi_1, Phi_2, Phi_3, Phi_4, Phi_5
                The formula for the matrix is:
                    u11 = cos[Theta_1]*cos[Theta_2]*exp[i*Phi_1]
                    u12 = sin[Theta_1]*exp[i*Phi_3]
                    u13 = cos[Theta_1]*sin[Theta_2]*exp[i*Phi_4]
                    u21 = sin[Theta_2]*sin[Theta_3]*exp[-i*Phi_4 - i*Phi_5] -
                            sin[Theta_1]*cos[Theta_2]*cos[Theta_3]*exp[i*Phi_1+i*Phi_2-i*Phi_3]
                    u22 = cos[Theta_1]*cos[Theta_3]*exp[i*Phi_2]
                    u23 = -cos[Theta_2]*sin[Theta_3]*exp[-i*Phi_1 - i*Phi_5] -
                            sin[Theta_1]*sin[Theta_2]*cos[Theta_3]*exp[i*Phi_2 - i*Phi_3 + i*Phi_4]
                    u31 = -sin[Theta_1]*cos[Theta_2]*sin[Theta_3]*exp[i*Phi_1 - i*Phi_3 + i*Phi_5]
                            - sin[Theta_2]*cos[Theta_3]*exp[-i*Phi_2-i*Phi_4]
                    u32 = cos[Theta_1]*sin[Theta_3]*exp[i*Phi_5]
                    u33 = cos[Theta_2]*cos[Theta_3]*exp[-i*Phi_1 - i*Phi_2] -
                            sin[Theta_1]*sin[Theta_2]*sin[Theta_3]*exp[-i*Phi_3 + i*Phi_4 + i*Phi_5]
    """

    # construct unitary, element by element
    u11 = np.cos(Theta_1)*np.cos(Theta_2)*np.exp(1j*Phi_1)
    u12 = np.sin(Theta_1)*np.exp(1j*Phi_3)
    u13 = np.cos(Theta_1)*np.sin(Theta_2)*np.exp(1j*Phi_4)
    u21 = np.sin(Theta_2)*np.sin(Theta_3)*np.exp(-1j*Phi_4 - 1j*Phi_5) - np.sin(Theta_1)*np.cos(Theta_2)*np.cos(Theta_3)*np.exp(1j*Phi_1+1j*Phi_2-1j*Phi_3)
    u22 = np.cos(Theta_1)*np.cos(Theta_3)*np.exp(1j*Phi_2)
    u23 = -1*np.cos(Theta_2)*np.sin(Theta_3)*np.exp(-1j*Phi_1 - 1j*Phi_5) - np.sin(Theta_1)*np.sin(Theta_2)*np.cos(Theta_3)*np.exp(1j*Phi_2 - 1j*Phi_3 + 1j*Phi_4)
    u31 = -1*np.sin(Theta_1)*np.cos(Theta_2)*np.sin(Theta_3)*np.exp(1j*Phi_1 - 1j*Phi_3 + 1j*Phi_5) - np.sin(Theta_2)*np.cos(Theta_3)*np.exp(-1j*Phi_2-1j*Phi_4)
    u32 = np.cos(Theta_1)*np.sin(Theta_3)*np.exp(1j*Phi_5)
    u33 = np.cos(Theta_2)*np.cos(Theta_3)*np.exp(-1j*Phi_1 - 1j*Phi_2) - np.sin(Theta_1)*np.sin(Theta_2)*np.sin(Theta_3)*np.exp(-1j*Phi_3 + 1j*Phi_4 + 1j*Phi_5)

    evaluated_unitary = np.array([[u11, u12, u13], [u21, u22, u23], [u31, u32, u33]])

    return evaluated_unitary

# based on old solovay kitaev code
def random_near_identity(n, alpha):
    # generate a random hermitian matrix
    H = np.array((np.random.rand(n,n) - 0.5) +1j * (np.random.rand(n,n) - 0.5))
    H = H + H.T.conjugate()
    # generate a unitary matrix from the hermitian matrix that is not far from the identity
    return np.array(sp.linalg.expm(1j * H * alpha))
    
def remap(U, order, d=2):
    U = np.array(U, dtype='complex128')

    swap = unitaries.general_swap(d)
    qudits = int(np.round(np.log(np.shape(U)[0]) / np.log(d)))
    beforemat = np.array(np.eye(np.shape(U)[0]), dtype='complex128')
    aftermat  = np.array(np.eye(np.shape(U)[0]), dtype='complex128')
    I = np.array(np.eye(d), dtype = 'complex128')
    if qudits == 1:
        return U
    current_order = [i for i in range(0, qudits)]
    for i in range(0, qudits):
        if not order[i] == current_order[i]:
            target_loc = i
            current_loc = current_order.index(order[i])
            while not target_loc == current_loc:
                if current_loc > target_loc:
                    # perform the swap current_loc and current_loc - 1
                    swapmat = matrix_kron(*[I]*(current_loc-1), swap, *[I]*(qudits - current_loc - 1))
#                    print("I"*(current_loc-1) + "SS" + "I" *(qudits - current_loc - 1))
                    current_order[current_loc], current_order[current_loc - 1] = current_order[current_loc - 1], current_order[current_loc]
                    beforemat = np.dot(beforemat, swapmat)
                    aftermat  = np.dot(swapmat, aftermat)
                    current_loc = current_loc - 1
                else:
                    # perform the swap current_loc and current_loc + 1
                    swapmat = matrix_kron(*[I]*(current_loc), swap, *[I]*(qudits - current_loc - 2))
 #                   print("I"*(current_loc) + "SS" + "I" *(qudits - current_loc - 2))
                    current_order[current_loc], current_order[current_loc + 1] = current_order[current_loc + 1], current_order[current_loc]
                    beforemat = np.dot(beforemat, swapmat)
                    aftermat  = np.dot(swapmat, aftermat)
                    current_loc = current_loc + 1

    return matrix_product(beforemat, U, aftermat)

def upgrade_qudits(U, di=2, df=3):
    qudits = int(np.log(U.shape[0])/np.log(di))
    new_unitary = np.array(np.eye(df**qudits), dtype='complex128')
    for i in range(df**qudits):
        skip = False
        testi = i
        oi = 0
        for dit in range(qudits):
            if testi % df >= di:
                skip = True
                break
            else:
                oi += (testi % df) *di**dit
            testi //= df

        if not skip:
            for j in range(df**qudits):
                skip = False
                testj = j
                oj = 0
                for dit in range(qudits):
                    if testj % df >= di:
                        skip = True
                        break
                    else:
                        oj += (testj % df) * di**dit
                    testj //= df

                if not skip:
                    new_unitary[i][j] = U[oi][oj]
    return new_unitary


def endian_reverse(U, d=2):
    n = int(np.log(U.shape[0])/np.log(d))
    return remap(U, list(reversed(range(0, n))), d)

def mpi_rank():
    if MPI is None:
        return 0 # allow fallback to multiprocessing
    comm = MPI.COMM_WORLD
    return comm.rank

def mpi_do_work(comm):
    """Do the work of a single compilation.

    Arguments:
        comm: An MPI communication object
    """
    done = False
    eval = None
    eval = comm.bcast(eval, root=0)
    while not done:
        done = comm.bcast(done, root=0)
        if done:
            break
        step = comm.recv(source=0, tag=comm.rank)
        if step is None:
            res = step # return None if we aren't doing work
        else:
            res = eval(step)
        data = comm.send(res, dest=0, tag=comm.rank)
        assert data is None

def mpi_worker():
    """Create a worker that will keep running compilation requests until told to stop"""
    # NOTE WELL: this should be kept in sync with the MPIParallelizer code in parallelizer.py
    if MPI is None:
        raise RuntimeError("MPI not installed")
    comm = MPI.COMM_WORLD
    done = False
    while not done:
        done = comm.bcast(done, root=0)
        if done:
            break
        mpi_do_work(comm)



"""
This module defines Parallelizer, which is a class that defines how to perform multiple circuits in parallel.

Several implementations are provided.

Attributes:
    LokyParallelizer : A Parallelizer based on Loky, a "deadlock-free" ProcessPoolExecutor
    MultiprocessingParallelizer : A Parallelizer based on multiprocessing
    ProcessPoolParallelizer : A Parallelizer based on concurrent.futures.ProcessPoolExecutor
    MPIParallelizer : A distributed MPI based Parallelizer
    SequentialParallelizer : Mostly for debugging purposes, a Parallelizer that runs tasks one at a time.
"""

from multiprocessing import get_context, cpu_count
from concurrent.futures import ProcessPoolExecutor
from functools import partial
import sys

try:
    from mpi4py import MPI
except ImportError:
    MPI = None

try:
    from loky import get_reusable_executor
except ImportError:
    get_reusable_executor = None


def default_num_tasks(options):
    return cpu_count()

def evaluate_step(tup, options):
    step, depth, weight = tup
    return (step, options.solver.solve_for_unitary(options.backend.prepare_circuit(step, options), options), depth, weight)

def single_task(opts):
    return 1

class Parallelizer():
    """Base class for all Parallelizers. Parallelizers calculate the value of multiple search nodes in parallel."""
    def solve_circuits_parallel(self, tuples):
        """Calculate the value of search tree nodes in parallel."""
        return None

    def done(self):
        """Finalize/Clean up any state needed to run the Parallelizer."""
        pass

class LokyParallelizer(Parallelizer):
    """A parallelizer based on Loky, a "deadlock-free" ProcessPoolExecutor.

    For more information on Loky see https://loky.readthedocs.io/en/stable/.
    """
    def __init__(self, options):
        options.set_smart_defaults(num_tasks=default_num_tasks)
        self.executor = get_reusable_executor(max_workers=options.num_tasks)
        self.process_func = partial(evaluate_step, options=options)

    def solve_circuits_parallel(self, tuples):
        return self.executor.map(self.process_func, tuples)

class MultiprocessingParallelizer(Parallelizer):
    """A Parallelizer based on muliprocessing. Note this cannot be used with the MultiStart_Solvers!"""
    def __init__(self, options):
        if sys.platform != 'win32':
            ctx = get_context('fork')
        else:
            ctx = get_context()
        options.set_smart_defaults(num_tasks=default_num_tasks)
        self.pool = ctx.Pool(options.num_tasks)
        self.process_func = partial(evaluate_step, options=options)

    def solve_circuits_parallel(self, tuples):
        yield from self.pool.imap_unordered(self.process_func, tuples)

    def done(self):
        self.pool.close()
        self.pool.terminate()
        self.pool.join()

class ProcessPoolParallelizer(Parallelizer):
    """A Parallelizer based on concurrent.futures.ProcessPoolExecutor."""
    def __init__(self, options):
        options.set_smart_defaults(num_tasks=default_num_tasks)
        if sys.version_info >= (3, 8, 0) and sys.platform != 'win32':
            ctx = get_context('fork')
            self.pool = ProcessPoolExecutor(options.num_tasks, mp_context=ctx)
        else:
            self.pool = ProcessPoolExecutor(options.num_tasks)

        self.process_func = partial(evaluate_step, options=options)

    def solve_circuits_parallel(self, tuples):
        return self.pool.map(self.process_func, tuples)

    def done(self):
        self.pool.shutdown()

class MPIParallelizer(Parallelizer):
    """A distributed MPI based Parallelizer.

    This implementation unfortunately requires some work on the part of the Project
    API or the user.
    """
    def __init__(self, options):
        options.set_smart_defaults(num_tasks=self.comm.size)
        if MPI is not None:
            self.comm = MPI.COMM_WORLD
            self.comm.bcast(False, root=0)
        else:
            raise RuntimeError("mpi4py is not installed")
        # HACK: We want to play nice with the multistart parallelizer
        #TODO this should probably be managed via the passed options going forward
        if hasattr(solver, 'num_threads'):
            self.tasks = self.comm.size - solver.num_threads
        else:
            self.tasks = self.comm.size
        eval = partial(evaluate_step, options)
        eval = self.comm.bcast(eval, root=0)

    def solve_circuits_parallel(self, tuples):
        # NOTE WELL: this should be kept in sync with the mpi_worker code in utils.py
        return self.map_steps(tuples)

    def map_steps(self, new_steps):
        base = 0
        done = False
        while len(new_steps) > 0:
            self.comm.bcast(False, root=0)
            for i in range(self.comm.size - 1):
                if i >= len(new_steps):
                    sendobj = None
                else:
                    sendobj = new_steps[i]
                self.comm.send(sendobj, dest=i+1, tag=i+1)
                res = self.comm.recv(source=i+1, tag=i+1)
                if res is not None:
                    yield res
            new_steps = new_steps[self.comm.size:]

    def done(self):
        self.comm.bcast(True, root=0)

class SequentialParallelizer(Parallelizer):
    """A Paralleizer that isn't, it runs tasks one at a time (mostly for debugging).
    """
    def __init__(self, options):
        options.set_smart_defaults(num_tasks=single_task)
        self.process_func = partial(evaluate_step, options=options)

    def solve_circuits_parallel(self, tuples):
        return map(self.process_func, tuples)



"""
A collection of constant gates and gate generators that are unusual or more complicated than those found in unitaries.py.
"""
from qsearch.gates import *
import scipy as sp

def generate_miro():
    """Generates a gate that was described as the X gate on the space of multiple qubits."""
    theta = np.pi/3
    c = np.cos(theta/2)
    s = -1j*np.sin(theta/2)
    return np.array([
        [c,0,0,0,0,0,0,s],
        [0,c,0,0,0,0,s,0],
        [0,0,c,0,0,s,0,0],
        [0,0,0,c,s,0,0,0],
        [0,0,0,s,c,0,0,0],
        [0,0,s,0,0,c,0,0],
        [0,s,0,0,0,0,c,0],
        [s,0,0,0,0,0,0,c]
        ], dtype='complex128')
mirogate = generate_miro()

def generate_HHL():
    def hadamard(theta: int = 0):
        return np.array([[np.cos(2*theta), np.sin(2*theta)],[np.sin(2*theta), -np.cos(2*theta)]])
    H = UGate(hadamard(), "H")
    RCH8 = CUGate(hadamard(np.pi/8), "H8", flipped=True)
    RCH16 = CUGate(hadamard(np.pi/16), "H16", flipped=True)
    RCY = CUGate(np.array([[0,-1j],[1j,0]]), "CY", flipped=True)
    RCNOT = UGate(np.array([[0,1,0,0],
                            [1,0,0,0],
                            [0,0,1,0],
                            [0,0,0,1]]), "RCNOT")
    SWAP = UGate(np.array([[1,0,0,0],
                            [0,0,1,0],
                            [0,1,0,0],
                            [0,0,0,1]]))

    # input parameters
    t0 = 2*np.pi
    A = np.array([[1.5, 0.5],[0.5, 1.5]])
    AU = sp.linalg.expm(1j * t0 * A / 2)

    CAU = CUGate(AU, "CA")
    CSH = CUGate(np.array([[1,0],[0,-1j]]), "CSH")

    X = UGate(np.array([[0,1],[1,0]]), "X")
    I = IdentityGate()

    circuit = ProductGate()
    circuit = circuit.appending(KroneckerGate(I,I,H))
    circuit = circuit.appending(KroneckerGate(I,CSH))
    circuit = circuit.appending(KroneckerGate(I,H,I))
    circuit = circuit.appending(KroneckerGate(I,I,H))
    circuit = circuit.appending(KroneckerGate(RCY, I))
    circuit = circuit.appending(KroneckerGate(SWAP,I))
    circuit = circuit.appending(KroneckerGate(I,RCY))
    circuit = circuit.appending(KroneckerGate(SWAP,I))
    return circuit.matrix([])
HHL = generate_HHL()




"""
This module defines the Gateset class, which represents the allowed gates and topology for a specific quantum computer.

Several Implementations of Gateset are also defined here.
Several aliases are also defined, for the most common use cases.

Attributes:
    ZXZXZCNOTLinear : A Gateset that uses CNOT and the ZXZXZ single qubit parameterization with the linear topology.
    U3CNOTLinear : A Gateset that uses CNOT and the U3 single qubit parameterization with the linear topology.
    QubitCNOTLinear : A Gateset that uses CNOT and the U3 single qubit parameterization with the linear topology, except it uses an XZXZ instead of a U3 after the control qubit of each CNOT.  This results in a gateset that covers the same search space as U3CNOTLinear, but with fewer redundant parameters, and therefore faster runtime.
    QubitCNOTRing : Uses U3 and XZXZ like QubitCNOTLinear, but includes a NonadjacentCNOTGate to add a link from the last qubit to the 0th.
    QubitCNOTAdjacencyList : Similar to QubitCNOTLinear and QubitCNOTRing, but takes in an adjacency list which uses NonadjacentCNOTGate to define work with a custom topology.
    QutritCPIPhaseLinear : A qutrit gateset that uses the CPIPhase gate as its two-qutrit gate, with a linear topology.
    QutritCNOTLinear : A qutrit gateset that uses an upgraded version of the CNOT gate as its two-qutrit gate, with a linear topology.

    DefaultQubit : The default Gateset for working with qubits.  Currently is equivalent to QubitCNOTLinear.
    DefaultQutrit : The default Gateset for working with qutrits.  Currently is equivalent to QutritCPIPhaseLinear.
    Default : The overall default Gateset, which is equivalent to DefaultQubit.
"""
from .gates import *
from .assemblers import flatten_intermediate
import numpy as np



class Gateset():
    """This class defines the supported gates and topology for a specific quantum hardware."""
    def __init__(self):
        """Gatesets must set the value of d in their initializer, which represents the size of qudits that are supported (e.g. 2 for qubits or 3 for qutrits)."""
        self.d = 0
        raise NotImplementedError("Gatesets must implemented their own initializers and must set self.d to reflect the size of the qudits implemented in the gateset")

    def initial_layer(self, qudits):
        """
        The initial layer in the compilation.  Usually a layer of parameterized single-qudit gates.

        Args:
            qudits : The number of qudits in this circuit.

        Returns:
            qsearch.gates.Gate : A single Gate representing an initial layer for the circuit
        """
        return None # NOTE: Returns A SINGLE gate

    # the set of possible multi-qubit gates for searching.  Generally a two-qubit gate with single qubit gates after it.
    def search_layers(self, qudits):
        """
        A set of possible multi-qubit gates for searching.  Usually this is a two-qudit gate followed by two single-qudit gates, for every allowed placement of the two-qudit gate.  This defines the branching factor of the search tree.

        Args:
            qudits : The number of qudits in this circuit
        
        Returns:
            list : A list of tuples of (gate,weight) where Gate is the Gate representing that possible placement of the two-qudit gate, and weight is the weight or cost of adding that gate in that placement to the final circuit.
        """
        return [] # NOTES: Returns a LIST of tuples of (gate, weight)

    def branching_factor(self, qudits):
        """
        Returns an integer indicating the expected branching factor.  Usually this is automatically determined from search_layers, but it may need to be overridden if successors is overridden.

        Args:
            qudits : The number of qudits in this circuit

        Returns:
            int : An integer indicating the expecte branching factor
        """
        # This is the default implementation, for Gatesets that rely on search_layers
        return len(self.search_layers(qudits))

    def successors(self, circ, qudits=None):
        """
        Returns a list of Gates that are successors in the search tree to the input Gate, circ, representing a current ansatz circuit.

        Args:
            circ : The curret ansatz Gate.
            qudits : The number of qudits in this circuit.

        Returns:
            list : A list of tuples of (gate, weight) where gate is a Gate that is a successor to circ, and weight is the cost or weight of moving to gate from circ.
        """

        # NOTE: Returns a LIST of tuples of (gate, weight)
        # NOTE: it is safe to assume that the circuit passed in here was produced by the functions of this class
        
        # This is the default implementation, for Gatesets that rely on search_layers
        qudits = int(np.log(circ.matrix([0]*circ.num_inputs).shape[0])/np.log(self.d))
        return [(circ.appending(t[0]), t[1]) for t in self.search_layers(qudits)]

    def __eq__(self, other):
        if self is other:
            return True
        if self.__module__ == Gateset.__module__:
            if type(self) == type(other):
                if self.__dict__ == other.__dict__:
                    return True
        return False

class ZXZXZCNOTLinear(Gateset):
    """A Gateset for working with CNOT and single-qubit gates parameterized with ZXZXZGate on the linear topology."""
    def __init__(self):
        self.single_gate = ZXZXZGate()
        self.cnot = CNOTGate()
        self.d = 2

    def initial_layer(self, n):
        return fill_row(self.single_gate, n)
    
    def search_layers(self, n):
        return linear_topology(self.cnot, self.single_gate, n, self.d)

class U3CNOTLinear(Gateset):
    """A Gateset for working with CNOT and single-qubit gates parameterized with U3Gate on the linear topology."""
    def __init__(self):
        self.single_gate = U3Gate()
        self.cnot = CNOTGate()
        self.d = 2

    def initial_layer(self, n):
        return fill_row(self.single_gate, n)
    
    def search_layers(self, n):
        return linear_topology(self.cnot, self.single_gate, n, self.d)

class QubitCNOTLinear(Gateset):
    """A Gateset for working with CNOT and single-qubit gates parameterized with U3Gate and XZXZGate on the linear topology.  This Gateset covers the same search space but uses fewer parameters than ZXZXZCNOTLinear and U3CNOTLinear."""
    def __init__(self):
        self.single_gate = U3Gate()
        self.single_alt  = XZXZGate()
        self.cnot = CNOTGate()
        self.d = 2

    def initial_layer(self, n):
        return fill_row(self.single_gate, n)
    
    def search_layers(self, n):
        return linear_topology(self.cnot, self.single_gate, n, self.d, single_alt=self.single_alt)

    def branching_factor(self, qudits):
        return qudits-1

    def successors(self, circ, qudits=None):
        if qudits is None:
            qudits = int(np.log(circ.matrix([0]*circ.num_inputs).shape[0])/np.log(self.d))
        skip_index = find_last_3_cnots_linear(circ)
        return [(circ.appending(layer[0]), layer[1]) for layer in linear_topology(self.cnot, self.single_gate, qudits, self.d, single_alt=self.single_alt, skip_index=skip_index)]

class QubitCNOTRing(Gateset):
    """A Gateset for working with CNOT and single-qubit gates parameterized with U3Gate and XZXZGate on the ring topology."""
    def __init__(self):
        self.single_gate = U3Gate()
        self.single_alt  = XZXZGate()
        self.cnot = CNOTGate()
        self.d = 2

    def initial_layer(self, n):
        return fill_row(self.single_gate, n)

    def search_layers(self, n):
        I = IdentityGate()
        gates = linear_topology(self.cnot, self.single_gate, n, self.d, identity_gate=I, single_alt=self.single_alt)
        if n == 2:
            return gates
        finisher = (ProductGate(NonadjacentCNOTGate(n, n-1, 0), KroneckerGate(self.single_gate, *[I]*(n-2), self.single_alt)), 1)
        return gates + [finisher]


class QubitCNOTAdjacencyList(Gateset):
    """A Gateset for working with CNOT and single-qubit gates parameterized with U3Gate and XZXZGate on a custom topology, specified in the initializer."""
    def __init__(self, adjacency):
        """
        Allows the specification of a custom topology through an adjacency list.

        For example, this is how you would specifiy the ring topology for 3 qubits:
        `[(0,1), (1,2), (2,1)]`

        It is not recommended to add bi-directional links, because with the arbitrary parameterized single qubit gates everywhere, such links would be redundant.

        Args:
            adjacency : A list of tuples specifying which CNOT placements are allowed.  The tuples must be in the form of (control, target).
        """
        self.single_gate = U3Gate()
        self.single_alt = XZXZGate()
        self.I = IdentityGate()
        self.adjacency = adjacency
        self.d = 2

    def initial_layer(self, n):
        return fill_row(self.single_gate, n)

    def search_layers(self, n):
        if n == 2:
            return [(ProductGate(CNOTGate(), KroneckerGate(self.single_gate, self.single_gate)), 1)] # prevents the creation of an extra cnot placement in the 2 qubit case

        gates = []
        for pair in self.adjacency:
            if pair[0] >= n or pair[1] >= n:
                continue
            cnot = NonadjacentCNOTGate(n, pair[0], pair[1])
            single_gates = [self.single_gate if i == pair[0] else self.single_alt if i == pair[1] else self.I for i in range(0, n)]
            gates.append((ProductGate(cnot, KroneckerGate(*single_gates)), 1)) 
        return gates

class QutritCPIPhaseLinear(Gateset):
    """A Gateset for working with CPIPhase and single-qutrit gates on the linear topology."""
    def __init__(self):
        self.single_gate = SingleQutritGate()
        self.d = 3

    def initial_layer(self, n):
        return fill_row(self.single_gate, n)
    
    def search_layers(self, n):
        return linear_topology(CPIPhaseGate(), self.single_gate, n, self.d)

class QutritCNOTLinear(Gateset):
    """A hybrid Gateset for working with CNOT and single-qutrit gates on the linear topology."""
    def __init__(self):
        self.single_gate = SingleQutritGate()
        self.d = 3

    def initial_layer(self, n):
        return fill_row(self.single_gate, n)
    
    def search_layers(self, n):
        return linear_topology(UpgradedConstantGate(CNOTGate()), self.single_gate, n, self.d)

# Commonly used functions for generating gatesets
def linear_topology(double_gate, single_gate, n, d, identity_gate=None, single_alt=None, double_weight=1, single_weight=0, skip_index=None):
    weight = double_weight + 2*single_weight
    if not identity_gate:
        identity_gate = IdentityGate(d=d)
    if single_alt is None:
        return [(KroneckerGate(*[identity_gate]*i, ProductGate(double_gate, KroneckerGate(single_gate, single_gate)), *[identity_gate]*(n-i-2)), weight) for i in range(0, n-1) if skip_index is None or i != skip_index]
    else:
        return [(KroneckerGate(*[identity_gate]*i, ProductGate(double_gate, KroneckerGate(single_alt, single_gate)), *[identity_gate]*(n-i-2)), weight) for i in range(0, n-1) if skip_index is None or i != skip_index]

def fill_row(gate, n):
    return KroneckerGate(*[gate]*n)


def find_last_3_cnots_linear(circuit):
    # this function finds the last 3 CNOTs in the circuit and returns the index if they are all in a row or None otherwise
    # this function is written specifically for linear topology-based gatesets.  Other gatesets should use find_last_3_cnots_arbitrary.
    # do not use either of these functions with gatesets not intended specifically for it

    contents = circuit._subgates
    if len(contents) < 4:
        return None
    indices = [[type(sub) for sub in gate._subgates].index(ProductGate) for gate in contents[-1:-4:-1]]

    if indices[0] == indices[1] and indices[1] == indices[2]:
        return indices[0]
    else:
        return None


# commonly used defaults
DefaultQubit = QubitCNOTLinear
DefaultQutrit = QutritCPIPhaseLinear
Default = DefaultQubit




try:
    from qiskit import QuantumCircuit
    import qiskit
except ImportError:
    qiskit = None
    raise ImportError("Cannot import qiskit, please run pip3 install qiskit before importing qiskit code.")

import numpy as np

from .gates import *


class QiskitImportError(Exception):
    """A class to represent issues importing code from qiskit"""


class QiskitGateConverter:
    def __init__(self, num_qubits):
        self.registers = []
        self.num_qubits = num_qubits
        self.parameters = []

    def convert(self, gate, qubits, cbits):
        """Abstraction to convert an arbitrary qiskit gate to a layer in a qsearch circuit"""
        if cbits != []:
            raise QiskitImportError("Classical operations are not supported in qsearch for now.")
        return getattr(self, f'convert_{gate.name}')(gate, qubits, cbits)
    
    def convert_cx(self, gate, qubits, cbits):
        for q in qubits:
            if q.register.name not in self.registers:
                if len(self.registers) == 0:
                    self.registers.append(q.register.name)
                else:
                    raise QiskitImportError("Qsearch does not support importing circuits with multiple quantum registers.")
        pair = [q.index for q in qubits]
        assert len(pair) == 2, "CNOT between more than 2 qubits?"
        return NonadjacentCNOTGate(self.num_qubits, pair[0], pair[1])

    def convert_u3(self, gate, qubits, cbits):
        assert len(qubits) == 1, "U3 on more than one qubit?"
        identity_gate = IdentityGate()
        self.parameters.extend(gate.params)
        index = qubits[0].index
        return KroneckerGate(*[identity_gate]*index, U3Gate(), *[identity_gate]*(self.num_qubits-index-1))

    def convert_u2(self, gate, qubits, cbits):
        assert len(qubits) == 1, "U2 on more than one qubit?"
        identity_gate = IdentityGate()
        self.parameters.extend(gate.params)
        index = qubits[0].index
        return KroneckerGate(*[identity_gate]*index, U2Gate(), *[identity_gate]*(self.num_qubits-index-1))

    def convert_rx(self, gate, qubits, cbits):
        assert len(qubits) == 1, "X on more than one qubit?"
        identity_gate = IdentityGate()
        self.parameters.extend(gate.params)
        index = qubits[0].index
        return KroneckerGate(*[identity_gate]*index, XGate(), *[identity_gate]*(self.num_qubits-index-1))

    def convert_ry(self, gate, qubits, cbits):
        assert len(qubits) == 1, "Y on more than one qubit?"
        identity_gate = IdentityGate()
        self.parameters.extend(gate.params)
        index = qubits[0].index
        return KroneckerGate(*[identity_gate]*index, YGate(), *[identity_gate]*(self.num_qubits-index-1))

    def convert_rz(self, gate, qubits, cbits):
        assert len(qubits) == 1, "Z on more than one qubit?"
        identity_gate = IdentityGate()
        self.parameters.extend(gate.params)
        index = qubits[0].index
        return KroneckerGate(*[identity_gate]*index, ZGate(), *[identity_gate]*(self.num_qubits-index-1))


def qiskit_to_qsearch(circ, converter=None):
    """Convert qiskit code to qsearch *structure* but not parameters"""
    converter = converter if converter is not None else QiskitGateConverter(circ.num_qubits)
    circuit = []
    for gate, qubits, cbits in circ.data:
        circuit.append(converter.convert(gate, qubits, cbits))
    return ProductGate(*circuit), np.array(converter.parameters)



"""
This module provides a wrapper that makes it easier to interface with the rest of Qsearch.
"""
import numpy as np
from numpy import matrix, array
from warnings import warn
from enum import Enum
import os
import shutil
import sys
import pickle
from multiprocessing import freeze_support
from .compiler import SearchCompiler
from . import solvers as scsolver
from .options import Options
from .defaults import standard_defaults, standard_smart_defaults
from . import logging, utils, gatesets, heuristics, assemblers
from time import time


try:
    from mpi4py import MPI
    from .utils import mpi_worker
except ImportError:
    MPI = None

class Project_Status(Enum):
    PROGRESS = 1
    COMPLETE = 2
    NOTBEGUN = 3

class Project:
    """The project class wraps most of the functionality of Qsearch as intended to help manage working with Qsearch."""
    def __init__(self, path, use_mpi=False):
        if MPI is not None and use_mpi:
            self.comm = MPI.COMM_WORLD
            self.rank = self.comm.rank
        else:
            self.comm = None
            self.rank = 0

        if self.rank == 0:
            self.folder = path
            self.name = os.path.basename(os.path.normpath(path))
            self.projfile = os.path.join(path, "qcproject")
            try:
                if not os.path.exists(path):
                    os.mkdir(path)
                with open(self.projfile, "rb") as projfile:
                    self._compilations, self.options = pickle.load(projfile)
                    self.set_defaults()
                    self.logger = logging.Logger(self.options.stdout_enabled, os.path.join(path, "{}-project-log.txt".format(self.name)), self.options.verbosity)
                    self.logger.logprint("Successfully loaded project {}".format(self.name))
                    self.status(logger=self.logger)
            except IOError:
                self._compilations = dict()
                self.options = Options()
                self.set_defaults()
                self.logger = logging.Logger(True, os.path.join(path, "{}-project-log.txt".format(self.name)), verbosity=1)

    def _save(self):
        with open(self.projfile, "wb") as projfile:
            pickle.dump((self._compilations, self.options), projfile)

    def _checkpoint_path(self, name):
        return os.path.join(self.folder, "{}.checkpoint".format(name))

    def add_compilation(self, name, U, options=None, handle_existing=None, **extraargs):
        """
        Adds a unitary to be compiled.

        Args:
            name : A name for this unitary.  Must be unique in this Project.
            U : The unitary to be compiled, in the form of a numpy ndarray with dtype="complex128"
            handle_existing : A variable which defines how to behave if a compilation with the given name already exists.  If it is set to "ignore", it will simply return without doing anything.  If it is set to "overwrite", it will overwrite the previous entry.  If it is set to the default of None, it will offer a warning asking the user to remove and re-add the compilation.
            options : The options passed to this function will be used only when this compilation is run.
            extraargs : The extraargs passed to this function will be used only when this compilation is run.
        """
        if name in self._compilations:
            s = self._compilation_status(name)
            if handle_existing == "ignore" or np.array_equal(U, self._compilations[name]["options"].target): # handle the case where this is an attempt to re-enter the same compilation
                if options is None and len(extraargs) == 0:
                    # a "noconflict" method for options could potentially let more cases fall through this return, thereby reducing the number of warnings
                    return
                elif s == Project_Status.PROGRESS or s == Project_Status.COMPLETE:
                    warn("Compilation {} already has made progress.  To change it, reset it first..".format(name), RuntimeWarning, stacklevel=2)
                    return
            elif handle_existing == "overwrite":
                self.remove_compilation(name)
            elif s == Project_Status.PROGRESS or s == Project_Status.COMPLETE:
                warn("A compilation with name {} already exists.  To change it, remove it and then add it again.".format(name), RuntimeWarning, stacklevel=2)
                return
        
        compopt = Options(statefile=self._checkpoint_path(name))
        compopt.update(options, **extraargs)
        compopt.target = U
        self._compilations[name] = {"options" : compopt}
        self._save()

    def __setitem__(self, keyword, value):
        try:
            if keyword in self.options and self.options[keyword] == value:
                self.options.update(**{keyword:value})
                return # no need to send out a warning if nothing is being changed
        except ValueError:
            self.options.update(keyword=value)
            return
        if not keyword in ["verbosity", "stdout_enabled"]: # "safe" keywords here
            for name in self._compilations:
                s = self._compilation_status(name)
                if s == Project_Status.COMPLETE or s == Project_Status.PROGRESS:
                    warn("This project contains compilations which have been completed or have been started.  Please call reset() to clear this progress before changing configurations.", RuntimeWarning, stacklevel=2)
                    return
        self.options.update(**{keyword:value})
        self._save()
        # adjust the logger if relevant
        if keyword == "verbosity":
            self.logger.verbosity = value
        elif keyword == "stdout_enabled":
            self.logger.std_enabled = value

    def configure_compiler_override(self, keyword, value):
        """An unsafe method that allows the user to set global Project Options even if there is existing work."""
        warn("Using this method could result in crashes, infinite loops, or other undefined behavior.  It is safer to reset the project and configure using project[\"keyword\"]=value.  Only use this method if the risk of error is worse than losing intermediate progress.")
        self.options.update(**{keyword:value})
        self._save()

    def __getitem__(self, keyword):
        return self.options[keyword]

    def __delitem__(self, keyword):
        del self.options[keyword]
        # reset logger properties to default if relevant
        if keyword == "verbosity":
            self.logger.verbosity = options.verbosity
        elif keyword == "stdout_enabled":
            self.logger.std_enabled = options.std_enabled

    def configure(self, **dictionary):
        """Adds multiple options to the global Project Options at once."""
        self.options.update(**dictionary)
 
    def reset(self, name=None):
        """Resets a Project, removing any work done but not the initial configurations.

        Args:
            name: Optionally specify a particular compilation by name to reset
        """
        if name is None:
            [self.reset(n) for n in self._compilations]
        else:
            cdict = self._compilations[name]
            self.get_options(name).checkpoint.delete()
            self._compilations[name] = {"options" : cdict["options"]}
        self._save()

    def remove_compilation(self, name):
        """Removes a compilation from a Project.

        Args:
            name: The name of the compilation to remove
        """
        warn("remove_compilation(name) is deprecated and will be removed in a future release.  Use clear(name) instead.", DeprecationWarning, stacklevel=2)
        self.get_options(name).checkpoint.delete()
        cdict = self._compilations.pop(name)
        self._save()

    def clear(self, name=None):
        """Clears a Project, reverting it to a state similar to a newly created Project.

        Args:
            name: Optionally specify a particular compilation by name to clear
        """
        if name is None:
            for name in self._compilations:
                self.get_options(name).checkpoint.delete()
            self._compilations = dict()
            self._compiler_config = dict()
        else:
            self.get_options(name).checkpoint.delete()
            cdict = self._compilations.pop(name)
        self._save()

    def __enter__(self):
        if self.rank == 0:
            return self
        else:
            mpi_worker()
            return self.__exit__(None, None, None)

    def __exit__(self, exc_typ, exc_val, exc_tb):
        if self.rank == 0:
            self.finish()
            return False
        else:
            sys.exit(0)

    def set_defaults(self):
        """Updates the Project Options with the standard defaults from defaults.py"""
        self.options.set_defaults(verbosity=1,stdout_enabled=True,blas_threads=None,compiler_class=SearchCompiler,**standard_defaults)
        self.options.set_smart_defaults(**standard_smart_defaults)

    def run(self):
        """Runs all of the compilations in the Project."""
        freeze_support()

        self.logger.logprint("Started running project {}".format(self.name))
        self.status(logger=self.logger)
        for name in self._compilations:
            cdict = self._compilations[name]

            runopt = self.get_options(name)
            blas_threads = runopt.blas_threads
            CompilerClass = runopt.compiler_class
            compiler = CompilerClass(runopt)

            if self._compilation_status(name) == Project_Status.COMPLETE:
                continue
            sublogger = logging.Logger(runopt.stdout_enabled, os.path.join(self.folder, "{}-log.txt".format(name)), runopt.verbosity)
            runopt.logger = sublogger
            self.logger.logprint("Starting compilation of {}".format(name))
            try:
                from threadpoolctl import threadpool_limits
            except ImportError:
                starttime = time()
                result = compiler.compile(runopt)
            else:
                with threadpool_limits(limits=blas_threads, user_api='blas'):
                    starttime = time()
                    result = compiler.compile(runopt)
            endtime = time()
            self.logger.logprint("Finished compilation of {}".format(name))
            cdict.update(**result)
            cdict["time"] = endtime - starttime
            self._compilations[name] = cdict
            self._save()
            self.logger.logprint("Recorded results from compilation.", verbosity=2)

            runopt.checkpoint.delete()
            self.logger.logprint("Deleted checkpoint file.", verbosity=2)
            self.status(logger=self.logger)
        self.logger.logprint("Finished running project {}".format(self.name))

    def post_process(self, postprocessor, name=None, options=None, **xtraargs):
        """Post-processes the specified compilation, or all compilations if name is None, using the specified postprocessor.

        Args:
            postprocessor: The qsearch.post_processing.PostProcessor to run on the compilation or project
            name : Optionally specify a particular compilation by name to reset
            options : Options to pass to the qsearch.post_processing.PostProcessor passed in `postprocessor`
            extraargs : Extra arguments passed as options to the qsearch.post_processing.PostProcessor passed in `postprocessor`
        """
        names = [name] if name else self._compilations
        for name in names:
            self.logger.logprint("Started postprocessing of {}".format(name))
            cdict = self._compilations[name]
            finaloptions = self.options.updated(cdict["options"]).updated(options, **xtraargs)
            result = postprocessor.post_process_circuit(cdict, finaloptions)
            cdict.update(**result)
            self._compilations[name] = cdict
            self.logger.logprint("Finished postprocessing of {}".format(name))
        self._save()
            
    def complete(self):
        """Returns a True if all compilations in the Project have finished and False otherwise."""
        return self._overall_status() == Project_Status.COMPLETE

    def finish(self):
        """Called when done running compilations in order to end MPI tasks."""
        if MPI is not None and self.comm is not None:
            self.comm.bcast(True, root=0)

    def status(self, name=None, logger=None):
        """Prints a status update on how much of a Project has finished running.

        Args:
            name: Optionally specify which compilation to check the status of
        """
        namelist = [name] if name else self._compilations
        for n in namelist:
            s = self._compilation_status(n)
            msg = ""
            if s == Project_Status.COMPLETE:
                msg = "Complete!"
            elif s == Project_Status.PROGRESS:
                msg = "In progress..."
            elif s == Project_Status.NOTBEGUN:
                msg = "Not started."
            elif s == Project_Status.DEBUGING:
                msg = "Debug."
            if logger is None:
                print("{}\t{}".format(n,msg))
            else:
                logger.logprint("{}\t{}".format(n,msg))

    @property
    def compilations(self):
        """The list of names corresponding to compilations on this Project."""
        return list(self._compilations.keys())

    def _compilation_status(self, name):
        cdict = self._compilations[name]
        if os.path.exists(self._checkpoint_path(name)):
            return Project_Status.PROGRESS
        elif "structure" in cdict and "parameters" in cdict:
            return Project_Status.COMPLETE
        else:
            return Project_Status.NOTBEGUN

    def _overall_status(self):
        complete = True
        started = False
        if name:
            return self._compilation_status(name)
        for name in self._compilations:
            s = self._compilation_status(name)
            if s != Project_Status.COMPLETE:
                complete = False
            if s != Project_Status.NOTBEGUN:
                started = True
        if complete:
            return Project_Status.COMPLETE
        elif started:
            return Project_Status.PROGRESS
        else:
            return Project_Status.NOTBEGUN

    def get_result(self, name):
        """Get the result of a compilation.

        Args:
            name: The name of the compilation to get the result dictionary from

        Returns:
            dict: The result dictionary for a finished compilation.  Usually this contains the entries "structure", a Gate, and "parameters", an array of real number parameters.
        """
        cdict = self._compilations[name]
        if not "structure" in cdict or not "parameters" in cdict:
            print("this compilation has not been completed.  please run the project to complete the compilation.")
        return cdict

    def get_target(self, name):
        """Get the target unitary of a compilation.

        Args:
            name: The name of the compilation to get the target from

        Returns:
            np.ndarray: The target unitary of the compilation
        """
        cdict = self._compilations[name]
        return cdict["options"].target

    def get_time(self, name):
        """Get the runtime that it took to run a compilation.

        Args:
            name: The name of the compilation to get the runtime of

        Returns:
            float: The number of seconds the compilation took
        """
        cdict = self._compilations[name]
        if "time" in cdict:
            return cdict["time"]
        else:
            return None

    def get_options(self, name=None):
        """Get the qsearch.options.Options object from a compilation of project

        Args:
            name: Optionally pass the name of the compilation to get the qsearch.options.Options object from

        Returns:
            qsearch.options.Options: the requested options object
        """
        if name is None:
            return self.options
        else:
            return self.options.updated(self._compilations[name]["options"])

    def assemble(self, name, options=None, **xtraargs):
        """Assembles a compilation using the Assembler specified as assembler in the Options.
        Args:
            name: The compilation to assemble
            options: Contains the qsearch.assemblers.Assembler to use in assembly

        Returns:
            str: The resulting assembled code
        """
        options = self.options.updated(options, **xtraargs)
        cdict = self._compilations[name]
        if not "structure" in cdict or not "parameters" in cdict:
            #TODO change this to a logprint
            print("This compilation has not been completed.  please run the project to complete the compilation.")
            return None

        out = options.assembler.assemble(cdict, options)
        if options.write_location is not None:
            with open(options.write_location, "w") as wfile:
                wfile.write(out)
        else:
            return out

